#include "stdafx.h"
#include "cwt.h"
#include "InOut.h"
#include "Helpers.h"

typedef list<float> float_list;


/************************ wavelet classes ***********************/

/* decision between real and complex wavelets made through the classes wavelet_real
   and wavelet_complex, the different behaviour is implemented by virtual methods */


///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_fkt
///////////////////////////////////////////////////////////////////
//	virtual void wavelet_fkt::init() = 0;

	void wavelet_fkt::init_args(const char *arguments) {
		if(!isstringempty(arguments)) {
			printf("The %s wavelet doesn't expect any arguments.\n", namewave);
			exit(1);
		}
	};

//	virtual float wavelet_fkt::evaluate(float x) const = 0;

	float wavelet_fkt::evaluate_clip(float x) const {   // evaluation with all values outside the support set to zero
		if( (lsupp <= x)  &&  (x <= rsupp) ) {
			return evaluate(x);
		} else {
			return 0;
		}
	};

//	virtual void wavelet_fkt::transform(const signal_fkt *sf, division_range *adiv, const arguments &arguments) = 0;
//	virtual void wavelet_fkt::discretize(float *w, long lpos, long rpos, float xc, float yc) const;

	void wavelet_fkt::get_supp (float as, long &lpos, long &rpos) const {
		if (as>0) {
			lpos = long(floor(lsupp * as));
			rpos = long( ceil(rsupp * as));
		} else {
			lpos = long(floor(rsupp * as));
			rpos = long( ceil(lsupp * as));
		}
	}

	void wavelet_fkt::get_max_supp (const arguments &arguments, long &sparebegin, long &spareend) const {
		if (arguments.a.min>0) {
			get_supp (fmax(arguments.a.min,arguments.a.max)*arguments.samplerate, sparebegin, spareend);
		} else {
			get_supp (fmin(arguments.a.min,arguments.a.max)*arguments.samplerate, sparebegin, spareend);
		}
	}

	// w points into the buffer at the position zero,
	//     to the left there must be at least -lpos
	// and to the right (including 0) there must be at least rpos words allocated
	void wavelet_fkt::discretize(float *w, long lpos, long rpos, float xc, float yc) const 
	{
		long k;
		w = w+lpos;
		for(k=lpos; k<rpos; k++) {
			*w = yc*evaluate_clip(float(k)*xc);
			//printf("%04ld: %g %g\n", k, (float(k)+0.5)/a, *w);
			w++;
		}
	}
///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_real
///////////////////////////////////////////////////////////////////

void wavelet_real::transform(const signal_fkt *sf, division_range *adiv, const char *outputbasename, 
														 const arguments &arguments, CArray<FLOAT,FLOAT>* pWaveletCoeffizients) 
{
	sf->transform (this, adiv, outputbasename, arguments, NULL,NULL,pWaveletCoeffizients);

	float F;
	float N;
	long	i;
	long nab=arguments.a.n*arguments.b.n;
	for(i=0; i<nab; i++) 
	{
		F=pWaveletCoeffizients->GetAt(i);
		N=sqrt(F*F);  
		pWaveletCoeffizients->SetAt(i,N);
	}//for


	convert_dat2pgm (outputbasename, this, arguments,pWaveletCoeffizients);
};

///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_haar
///////////////////////////////////////////////////////////////////

	void wavelet_haar::init() {
		cpsi=sqrt(2.0); lsupp=-0.5; rsupp=0.5;
		strcpy(namewave,"Haar");
	}

	float wavelet_haar::evaluate(float x) const 
	{
		if(x<0) 
		{
			return -1;
		} 
		else 
		{
			return 1;
		}
	}

///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_mexicanhat
///////////////////////////////////////////////////////////////////

	void wavelet_mexicanhat::init() 
	{
		cpsi=sqrt(6.0); lsupp=-5; rsupp=5;
		strcpy(namewave,"Mexican hat"); 
	}
	
	float wavelet_mexicanhat::evaluate(float x) const 
	{
		x *= sqrt(M_PI)*3/8;
		float x2 = x*x;
		return ( 1 - x2 ) / exp(x2/2);
	}

///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_sinus
///////////////////////////////////////////////////////////////////

	void wavelet_sinus::init() 
	{
		cpsi=1; lsupp=-0.5; rsupp=0.5;
		strcpy(namewave,"Sinus"); 
	}
	float wavelet_sinus::evaluate(float x) const 
	{
		return sin(2*M_PI*x);
	}

///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_poisson
///////////////////////////////////////////////////////////////////
	void wavelet_poisson::init() 
	{
		cpsi=1; lsupp=-10; rsupp=10;
		strcpy(namewave,"Poisson");
	}

	float wavelet_poisson::evaluate(float x) const 
	{
		// according to Holschneider: Wavelets; an analysis tool
		// g(t) = t*P'(t)+P(t), with P(t) = 1/(pi*(1+t^2))
		// its fourier representation is according to Holschneider (verified with mathematica)
		// g^(w) = abs(w) * exp(-abs(w))
		x = x*2/3;   // move center frequency to 1
		float x2 = x*x;
		return (1-x2)/(M_PI*(1+x2));
	}

///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_complex
///////////////////////////////////////////////////////////////////

void wavelet_complex::base_transform(const signal_fkt *sf, division_range *adiv, const char *outputbasename,  
																		 const arguments &arguments, CArray<FLOAT,FLOAT>* pWaveletCoeffizientsReal,
																		 CArray<FLOAT,FLOAT>* pWaveletCoeffizientsImg,
																		 CArray<FLOAT,FLOAT>* pWaveletCoeffizients) 
{
	safe_string complexbasename(strlen(outputbasename)+3);

	complexbasename.append(outputbasename);
	complexbasename.append(".re");
	cc = cc_real;
	sf->transform (this, adiv, complexbasename.cstr(), arguments, pWaveletCoeffizientsReal,
									pWaveletCoeffizientsImg, pWaveletCoeffizients);
	if (!sf->complex_valued) //wenn Signal nicht Komplex
	{
		pWaveletCoeffizientsReal->Copy(*pWaveletCoeffizients);
		pWaveletCoeffizients->RemoveAll();
	}
	complexbasename.clear();
	complexbasename.append(outputbasename);
	complexbasename.append(".im");
	cc = cc_imag;
	sf->transform (this, adiv, complexbasename.cstr(), arguments, pWaveletCoeffizientsReal,
									pWaveletCoeffizientsImg,pWaveletCoeffizients);
	if (!sf->complex_valued) //wenn Signal nicht Komplex
	{
		pWaveletCoeffizientsImg->Copy(*pWaveletCoeffizients);
		pWaveletCoeffizients->RemoveAll();
	}

//	sf->merge_complex (this, arguments, outputbasename, ".re", ".re.re", ".im.im", sub, pWaveletCoeffizientsReal);
//	sf->merge_complex (this, arguments, outputbasename, ".im", ".re.im", ".im.re", add, pWaveletCoeffizientsImg);
}

void wavelet_complex::transform(const signal_fkt *sf, division_range *adiv, const char *outputbasename, 
																const arguments &arguments,CArray<FLOAT,FLOAT>* pWaveletCoeffizients) 
{
	CArray<FLOAT,FLOAT> WaveletCoeffizientsReal;
	CArray<FLOAT,FLOAT> WaveletCoeffizientsImg;
	
	WaveletCoeffizientsReal.RemoveAll();
	WaveletCoeffizientsImg.RemoveAll();

	base_transform (sf, adiv, outputbasename, arguments,&WaveletCoeffizientsReal,&WaveletCoeffizientsImg,
									pWaveletCoeffizients);

	/* merge the results of both real transforms */
	complex_wt (outputbasename, this, arguments,&WaveletCoeffizientsReal,
							&WaveletCoeffizientsImg,pWaveletCoeffizients);
};
///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_morlet
///////////////////////////////////////////////////////////////////
	void wavelet_morlet::init() 
	{
		cpsi=sqrt(M_PI); rsupp=2*M_PI; lsupp=-rsupp;
		strcpy(namewave,"Morlet");
		cc = cc_real;
	}

	void wavelet_morlet::init_args(const char *arguments) 
	{
		scale = 1;
		if(!isstringempty(arguments)) {
			char *end;
			if(!wavargtof(arguments,end,scale) || *end!=0) {
				printf("Morlet wavelet accepts only one float argument for frequency, but not <%s>\n", arguments+1);
				exit(1);
			}
		}
		rsupp=2*M_PI*scale; lsupp=-rsupp;
		//printf("morlet scale %g\n", scale);
	}

	float wavelet_morlet::evaluate(float x) const 
	{
		// this condition has to be checked for every evaluation of function values which is very inefficient
		float envx = x/scale;
		if (cc == cc_real) {
			return cos(2*M_PI*x) * exp(-envx*envx/2);
		} else {
			return sin(2*M_PI*x) * exp(-envx*envx/2);
		}
	}

///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_cauchy
///////////////////////////////////////////////////////////////////

	void wavelet_cauchy::init() 
	{
		cpsi=1; lsupp=-5; rsupp=5;
		strcpy(namewave,"Cauchy");
		cc = cc_real;
	}

	void wavelet_cauchy::init_args(const char *arguments) 
	{
		order = 1;
		if(!isstringempty(arguments)) {
			char *end;
			if(!wavargtof(arguments,end,order) || *end!=0 || order < 0) {
				printf("Cauchy wavelet accepts only one non-negative argument for the order, but not <%s>\n", arguments+1);
				exit(1);
			}
		}
		cpsi = 1.0f/(2*M_PI);   // computing a sophisticated factor is not that easy
		freq = 4.0*M_PI/float(2*order+1); // looks good, but I expected that we must divide by 2*M_PI ...
		// set support, where psi is above 1/20 of its maximum
		rsupp = pow(20,1.0/order)/freq;
		lsupp = -rsupp;
		printf("Cauchy: lsupp %g, rsupp %g\n", lsupp, rsupp);
	}

	float wavelet_cauchy::evaluate(float x) const 
	{
		complex<float> t(1,-x*freq);
		complex<float> frac = pow(t,-order-1);

		if (cc == cc_real) {
			return real(frac);
		} else {
			return imag(frac);
		}
	}

///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_spinning
///////////////////////////////////////////////////////////////////

// cauchy wavelet where you can specify the variation in scale and time direction
	void wavelet_spinning::init() 
	{
		cpsi=1; lsupp=-20; rsupp=20;
		strcpy(namewave,"Spinning");
	}

	void wavelet_spinning::init_args(const char *arguments) 
	{
		float mu1, mu2;
		if(isstringempty(arguments)) {
			printf("Two arguments required for spinning wavelet.\n");
			exit(1);
		} else {
		char *end;
		if(!wavargtof(arguments,end,mu1) || !wavargtof(end,end,mu2) || *end!=0) {
				printf("Spinning wavelet requires two arguments but not <%s>\n", arguments+1);
				exit(1);
			}
		}
		printf("Spinning: mu1 %g, mu2 %g\n", mu1, mu2);
		if(mu2 < 0.25) {
			printf("To get an admissible wavelet it must hold mu2 >= 1/4, but for mu2=%f this does not hold.\n", mu2);
			exit(1);
		}
		float sqrtmu2 = sqrt(mu2);
		alpha = complex<float>(-0.5-sqrtmu2,mu1);
		freq = 2*M_PI/sqrtmu2;
		// computing a sophisticated value for 'cpsi' is not that easy and requires evaluation of the Gamma function
		// set support, where psi is above 1/tol of its maximum
		const float &ar = real(alpha);
		const float &ai = imag(alpha);
		float tol = log(20.0) - (ar*log(1+ai*ai)/2 + ai*atan(ai));
		float loglsupp, logrsupp;
		if (ai > 0) {
			loglsupp = (tol     ) / -ar;
			logrsupp = (tol + ai) / -ar;
		} else {
			loglsupp = (tol - ai) / -ar;
			logrsupp = (tol     ) / -ar;
		};
		lsupp = -exp(loglsupp)/freq;
		rsupp =  exp(logrsupp)/freq;
		//printf("Support of cauchy of same order: supp %g\n", pow(20,-1.0/ar)/freq);
		printf("Spinning: lsupp %g, rsupp %g\n", lsupp, rsupp);
	}

	float wavelet_spinning::evaluate(float x) const 
	{
		complex<float> t(1,-x*freq);
		complex<float> frac = pow(t,alpha);

		if (cc == cc_real) {
			return real(frac);
		} else {
			return imag(frac);
		}
	}


///////////////////////////////////////////////////////////////////
//Memberfunktionen von wavelet_bspline
///////////////////////////////////////////////////////////////////
void wavelet_bspline::reset()
{
	pwidths->clear();
	delete pwidths;
//	widths.RemoveAll();
}

enum {bsm_simple, bsm_accumulate, bsm_fractional} bsplinetransform_modi;

	void wavelet_bspline::init() 
	{
		cpsi=1; lsupp=-1; rsupp=1;
		strcpy(namewave,"B-Spline");
		cc = cc_real;
		pwidths = new float_list;
	}

	void wavelet_bspline::init_args(const char *arguments) 
	{
		rsupp = 0;
		while(!isstringempty(arguments)) {
			float width;
			char* end;
			if(!wavargtof(arguments,end,width) || width <= 0) {
				printf("B-Spline wavelet accepts a comma separated list of positive floats, but not <%s>\n", arguments+1);
			exit(1);
			}
			rsupp += width;
			pwidths->push_back(width);
//			widths.AddTail(width);
			arguments = end;
		}
		if (pwidths->empty()) {
//		if (widths.IsEmpty()) {
			printf("At least one window width argument is required for the B-spline wavelet.\n");
			// otherwise the support is [0,0] which is too small
			exit(1);
		}
		rsupp = rsupp/2;
		lsupp = -rsupp;
		printf("B-Spline: lsupp %g, rsupp %g\n", lsupp, rsupp);
	}

	float wavelet_bspline::evaluate(float x) const 
	{
		printf("wavelet_bspline::evaluate is not implemented - use wavelet_bspline::discretize instead\n");
		exit(1);
		return 0;
	}
	// similar to the global 'smooth' routine
	// buffer z must contain at least len+n-1 words
	void wavelet_bspline::wide_smooth(float *z, const float *ya, long len, long n) const 
	{
		float acc=0;
		const float *yl = ya;
		const float *yr = ya;
		float nf = 1.0/float(n);
		long i,j;
		for ((i=1-n,j=0);i<len;(z++,i++,j++)) {
			if(i >   0) {acc -= *yl; yl++; }
			if(j < len) {acc += *yr; yr++; }
			*z = acc*nf;
		}
	}

	void wavelet_bspline::discretize(float *w, long lpos, long rpos, float xc, float yc) const 
	{
		long yn = 1;
		safe_signal y(1);
		y.cptr()[0] = 1;
		float_list::const_iterator it;
//		POSITION it;
//		int k=0;
		for(it = pwidths->begin(); it != pwidths->end(); it++) {
//		for(it = widths.GetHeadPosition(); it != widths.GetTailPosition(); it++) {
			float width = *it / fabs(xc);
//			float width = widths.GetAt(it) / fabs(xc);
			long n = long(ceil(width));
			safe_signal z(yn+(n-1));
			wide_smooth (z.cptr(), y.cptr(), yn, n);
			y.replace(z.untied_cptr());
			yn += (n-1);
//			char debug_name[100];
//			sprintf(debug_name,"c://Dumps//debug_bspline%d.txt", k++);
//			write_signal(debug_name,y.cptr(),yn);
		}
		long n=yn/2;
		//printf("lpos %ld, rpos %ld, yn %ld, n %ld\n", lpos, rpos, yn, n);
		long i, ie;
		ie = min(rpos,-n);
		for(i=lpos; i<=ie; i++) {
			w[i] = 0;
		}
		ie = min(rpos,n);
		{
			const float *z = y.cptr()+(n-1);
			if (cc == cc_real) {
				for(; i<ie; i++) {
					w[i] = z[i] * cos(float(i)*xc*2*M_PI);
				}
			} else {
				for(; i<ie; i++) {
					w[i] = z[i] * sin(float(i)*xc*2*M_PI);
				}
			}
		}
		for(; i<rpos; i++) {
			w[i] = 0;
		}
//		write_signal("c://Dumps//debug_bspline_sin.txt",w+lpos,rpos-lpos);
	}

	// Swapping of 'const' buffers can't be used for non-constant buffers - why says gcc that?
	inline void wavelet_bspline::swap_buffer(float *&a, float *&b) 
	{
		float *
		c = a;
		a = b;
		b = c;
	}

	inline void wavelet_bspline::swap_buffer(const float *&a, const float *&b) 
	{
		const float *
		c = a;
		a = b;
		b = c;
	}

void wavelet_bspline::transform(const signal_fkt *sf, division_range *adiv, const char *outputbasename, 
																const arguments &arguments, CArray<FLOAT,FLOAT>* pWaveletCoeffizients) 
{
	CArray<FLOAT,FLOAT> WaveletCoeffizientsReal;
	CArray<FLOAT,FLOAT> WaveletCoeffizientsImg;
	
	WaveletCoeffizientsReal.RemoveAll();
	WaveletCoeffizientsImg.RemoveAll();

	if(!arguments.usesignalrate) {
		base_transform (sf, adiv, outputbasename, arguments, &WaveletCoeffizientsReal,&WaveletCoeffizientsImg,
										pWaveletCoeffizients);

	} else {
		fasttransform (sf, adiv, outputbasename, arguments, &WaveletCoeffizientsReal, &WaveletCoeffizientsImg,
										pWaveletCoeffizients);
/*		{
			safe_string complexbasename(strlen(outputbasename)+3);

			complexbasename.append(outputbasename);
			complexbasename.append(".re");
			convert_dat2pgm(complexbasename.cstr(), this, arguments, &WaveletCoeffizientsReal);

			complexbasename.clear();
			complexbasename.append(outputbasename);
			complexbasename.append(".im");
			convert_dat2pgm(complexbasename.cstr(), this, arguments,&WaveletCoeffizientsImg);
			}
*/
	}

	/* merge the results of both real transforms */
	complex_wt (outputbasename, this, arguments,&WaveletCoeffizientsReal,&WaveletCoeffizientsImg,pWaveletCoeffizients);
	convert_dat2pgm("c://Dumps//Rest", this, arguments,pWaveletCoeffizients);

};


void wavelet_bspline::fasttransform(const signal_fkt *sf, division_range *adiv, const char *outputbasename,
																		const arguments &arguments, 
																		CArray<FLOAT,FLOAT>* pWaveletCoeffizientsReal,
																		CArray<FLOAT,FLOAT>* pWaveletCoeffizientsImg,
																		CArray<FLOAT,FLOAT>* pWaveletCoeffizients)
{
	long lmaxpos, rmaxpos;
	get_max_supp (arguments, lmaxpos, rmaxpos);
	long size = arguments.bmaxint-arguments.bminint;
	long totalsize = size+rmaxpos-lmaxpos;
	safe_signal ws0r(totalsize), ws0i(totalsize);
	safe_signal ws1r(totalsize), ws1i(totalsize);
	// ***************** dangerous --- deallocation might be skipped by an exception
	const float *sr  = sf->discretize(arguments,this,cc_real);
	const float *si  = sf->discretize(arguments,this,cc_imag);

//	safe_file rfile(wfopen(outputbasename,".re.dat","w"));
//	safe_file ifile(wfopen(outputbasename,".im.dat","w"));
//	printhbar();
//	printf("   compute the fast b-spline wavelet transform ...\n ");
//	arguments.print_wtinfo();
	time_t cta = time(NULL);
	float bstep = float(arguments.bmaxint-arguments.bminint)/arguments.b.n;
	long hna;

	pWaveletCoeffizientsReal->SetSize(arguments.a.n*arguments.b.n);
	pWaveletCoeffizientsImg->SetSize(arguments.a.n*arguments.b.n);

	for(hna = arguments.a.n; hna>0; hna--) {
		float a = adiv->get_nth(hna);
		float as = adiv->get_nth(hna)*arguments.samplerate;
		float absas = fabs(as);
		float omega = 2*M_PI/as;
		const float *init_si;
		if (si == NULL) { // real signal
			ws0i.set(totalsize,0);
			init_si = ws0i.cptr();
		} else { // complex signal
			init_si = si;
		}
		fast_modulate(ws1r.cptr(),ws1i.cptr(),sr,init_si,totalsize,omega*lmaxpos,omega);

		long lsuppint = 0;
		float_list::const_iterator it;
//		POSITION it;
		// several alternative implementation with different smooth images
		switch (bsm_accumulate) {
			case bsm_simple:
				for (it = pwidths->begin(); it != pwidths->end(); it++) {
//				for (it = widths.GetHeadPosition(); it != widths.GetTailPosition(); it++) {
					long winwid = long(ceil(*it*absas));
//					long winwid = long(ceil(widths.GetAt(it)*absas));
					lsuppint -= winwid;
					smooth(ws0r.cptr(),ws1r.cptr(),totalsize-winwid,winwid); ws0r.swap(ws1r);
					smooth(ws0i.cptr(),ws1i.cptr(),totalsize-winwid,winwid); ws0i.swap(ws1i);
				}
				lsuppint /= 2;
				break;
			case bsm_accumulate:
				// reduce rounding errors caused by rounding each width separately
				// this is more precise, but the exact result can't be predicted
				// and we are still restricted to integer values of the b-spline support
				{
					float totalwidth = 0;
					long oldtotalwidth = 0;
					for (it = pwidths->begin(); it != pwidths->end(); it++) {
//					for (it = widths.GetHeadPosition(); it != widths.GetTailPosition(); it++) {
						totalwidth += *it;
//						totalwidth += widths.GetAt(it);
						long newtotalwidth = long(ceil(totalwidth*absas));
						long winwid = newtotalwidth - oldtotalwidth;
						//printf("winwid %ld  ", winwid);
						smooth(ws0r.cptr(),ws1r.cptr(),totalsize-winwid,winwid); ws0r.swap(ws1r);
						smooth(ws0i.cptr(),ws1i.cptr(),totalsize-winwid,winwid); ws0i.swap(ws1i);
						oldtotalwidth = newtotalwidth;
					}
					//printf("\n");
					lsuppint += long(floor(lsupp*absas));
				}
				break;
			case bsm_fractional:
				// using this method the reserved spare border for the signal may be a bit too small
				for (it = pwidths->begin(); it != pwidths->end(); it++) {
//				for (it = widths.GetHeadPosition(); it != widths.GetTailPosition(); it++) {
					float wid2 = *it*absas/2;
//					float wid2 = widths.GetAt(it)*absas/2;
					long winwid2 = long(ceil(wid2));
					lsuppint -= winwid2;
					fract_sym_smooth(ws0r.cptr(),ws1r.cptr()+winwid2,totalsize-2*winwid2,wid2); ws0r.swap(ws1r);
					fract_sym_smooth(ws0i.cptr(),ws1i.cptr()+winwid2,totalsize-2*winwid2,wid2); ws0i.swap(ws1i);
				}
				break;
		}
		// by smoothing we have moved signal to the right, that's why another modulation offset is chosen for demodulation
		fast_modulate(ws0r.cptr(),ws0i.cptr(),ws1r.cptr(),ws1i.cptr(),totalsize,-omega*(lmaxpos-lsuppint),-omega);

		long hnb;
		float b;
		float yc=sqrt(fabs(a));
		float *wsr = ws0r.cptr()-lmaxpos+lsuppint;
		float *wsi = ws0i.cptr()-lmaxpos+lsuppint;
		//printf("lmaxpos %ld, long(a*lsupp) %ld, totalsize %ld\n", lmaxpos, long(a*lsupp), totalsize);
		for((hnb = arguments.b.n, b = 0); hnb>0; (hnb--, b+=bstep)) {
				pWaveletCoeffizientsReal->SetAt((arguments.a.n-hna)*arguments.b.n+(arguments.b.n-hnb),wsr[long(b)]*yc); //Realteil
				pWaveletCoeffizientsImg->SetAt((arguments.a.n-hna)*arguments.b.n+(arguments.b.n-hnb),wsi[long(b)]*yc);	//Imaginrteil
//				pWaveletCoeffizientsReal->Add(wsr[long(b)]*yc); //Realteil
//				pWaveletCoeffizientsImg->Add(wsi[long(b)]*yc);	//Imaginrteil
//			fprintf(rfile.cfile(),"%g \n",wsr[long(b)]*yc);
//			fprintf(ifile.cfile(),"%g \n",wsi[long(b)]*yc);
   	}
	}
	sf->free_discretized(sr);
	sf->free_discretized(si);

/*	time_t ctb = time(NULL);
	CString Time,Text;
	sprintf(Time.GetBuffer(100),"%.2f", difftime(ctb, cta));
	AfxFormatString1(Text,IDS_FASTTRANSFORM,Time);
	AfxMessageBox(Text);
	*/
}



	/****************** signal classes **************************/

/* We have to perform different actions depending on whether the signal is real or complex
   and if the signal is discretized or continuous.
	We can't put these two aspects easily in a class structure,
	thus we will use sub classes for distinguishing between
	discretized and continuous signals. The decision between real and complex values
	has to be made by each routine separately.
*/

///////////////////////////////////////////////////////////////////
//Memberfunktionen von signal_fkt
///////////////////////////////////////////////////////////////////

	float signal_fkt::evaluate_imag(float x) const 
	{
		printf("Program error: evaluate_imag() implemented only for complex valued signals. Please check signal_fkt::complex_valued before calling evaluate_imag()!\n");
		exit(1);

	};

void signal_fkt::transform(const wavelet_fkt *wf, division_range *adiv, const char *outputbasename, 
													 const arguments &arguments,
													 CArray<FLOAT,FLOAT>* pWaveletCoeffizientsReal,
													 CArray<FLOAT,FLOAT>* pWaveletCoeffizientsImg,
													 CArray<FLOAT,FLOAT>* pWaveletCoeffizients) const //Proz. zur Berechnung der WT
{
	if (!complex_valued) {
		wavelettransform (this, cc_real, wf, adiv, outputbasename, arguments, pWaveletCoeffizients);

	} else {
		safe_string complexbasename(strlen(outputbasename)+3);
		complexbasename.append(outputbasename);
		complexbasename.append(".re");
		wavelettransform (this, cc_real, wf, adiv, complexbasename.cstr(), arguments, pWaveletCoeffizientsReal);

		complexbasename.clear();
		complexbasename.append(outputbasename);
		complexbasename.append(".im");
		wavelettransform (this, cc_imag, wf, adiv, complexbasename.cstr(), arguments, pWaveletCoeffizientsImg);
	}
	//aufrumen
}

// this will only be called from wavelet_complex::base_transform
void signal_fkt::merge_complex (const wavelet_complex *wf, const arguments &arguments, 
																const char *outputbasename, const char *outext, 
																const char *inaext, const char *inbext, superimpose_modes mode,
																CArray<FLOAT,FLOAT>* pWaveletCoeffizients) const 
{
	safe_string complexbasename(strlen(outputbasename)+3);
	complexbasename.append(outputbasename);
	complexbasename.append(outext);
	// join the separate products if we must compute the complex-complex transform
	if (complex_valued) {
		// zr = xr*yr - xi*yi
		safe_string aname(strlen(outputbasename)+6);
		safe_string bname(strlen(outputbasename)+6);
		aname.append(outputbasename); aname.append(inaext);
		bname.append(outputbasename); bname.append(inbext);
		superimpose_dat (complexbasename.cstr(), aname.cstr(), bname.cstr(), mode);
		delete_dat (aname.cstr());
		delete_dat (bname.cstr());
		printf("\n   merge into complex part %s\n", complexbasename.cstr());
	}
	convert_dat2pgm (complexbasename.cstr(), wf, arguments, pWaveletCoeffizients);
}



///////////////////////////////////////////////////////////////////
//Memberfunktionen von signal_continuous
///////////////////////////////////////////////////////////////////
	const float *signal_continuous::discretize(const arguments &args, const wavelet_fkt *wf, 
																											complex_component cc) const 
	{
		if (complex_valued || cc == cc_real) {
			long totalbegin, totalend, size;
			get_signalsupp (args, wf, totalbegin, totalend);
			size = totalend - totalbegin;
			safe_signal f(size);
			long i;
			if (cc==cc_imag) {
				for (i=0; i<size; i++) {
					f.cptr()[i] = evaluate_imag(float(i+totalbegin));
				}
			} else {
				for (i=0; i<size; i++) {
					f.cptr()[i] = evaluate_real(float(i+totalbegin));
				}
			}

			return f.untied_cptr();
		} else {
			return NULL;
		}
	};

	void signal_continuous::free_discretized(float *f) const 
	{
		if (f!=NULL) {
			delete[] f;
		}
	};

	void signal_continuous::write (FILE *file, const arguments &args, complex_component cc) const 
	{
/*		long i;
		long size  = abs(args.bmaxint - args.bminint);
		long begin = min(args.bmaxint, args.bminint);
		if (cc==cc_imag) {
			for (i=0; i<size; i++) {
				fprintf(file, "%f\n", evaluate_imag(float(i+begin)));
			}
		} else {
			for (i=0; i<size; i++) {
				fprintf(file, "%f\n", evaluate_real(float(i+begin)));
			}
		}
		*/
	};

///////////////////////////////////////////////////////////////////
//Memberfunktionen von signal_discrete
///////////////////////////////////////////////////////////////////

	signal_discrete::signal_discrete(SAMPLE* input, int input_length, const arguments &args, const wavelet_fkt *wf) 
	{  //Einlesen des Datensatzes
		get_signalsupp (args, wf, totalbegin, totalend);

		bool text_format = true;
		complex_valued = false;

		long size = totalend - totalbegin;
		printf("size %ld\n", size);

		// if totalend is greater than zero it holds always that skipdata - leadingszeros = totalbegin
		long leadingzeros = min(0,totalend)-min(0,totalbegin);
		long skipdata     =                 max(0,totalbegin);
		long i;
		printf("leadingzeros %ld, skipdata %ld\n", leadingzeros, skipdata);
		printhbar();

		safe_signal newfr(size);
		newfr.set(leadingzeros,0);

		safe_signal newfi;
		if (complex_valued) {
			newfi.replace(new float[size]);
			newfi.set(leadingzeros,0);
		}
		CString Ausgabe;
		sprintf(Ausgabe.GetBuffer(200)," reading the data set from %ld to %ld\n",totalbegin,totalend);
//		AfxMessageBox(Ausgabe);
		if (text_format) 
		{
			for(i=leadingzeros; i<size && i<input_length; i++) 
			{
				newfr.cptr()[i]=input[i];
   		}
		} 
		for(; i<size; i++) {
			newfr.cptr()[i]=0;
		}
		fralloc = newfr.untied_cptr();
		fr      = fralloc-totalbegin;
	};


//	signal_discrete::~signal_discrete () 
	signal_fkt::~signal_fkt() 
	{
		if (fralloc!=NULL) {
			delete[] fralloc;
		}
		if (fialloc!=NULL) {
			delete[] fialloc;
		}
	};

	const float *signal_discrete::discretize(const arguments &args, const wavelet_fkt *wf, complex_component cc) const 
	{
		if (cc == cc_imag) {
			return fialloc;
		} else {
			return fralloc;
		}
	};

	float signal_discrete::evaluate_real(float x) const 
	{
		long i = long(x);
		// if the padding with zeros was successful, the 'if' statement can be removed
		if(totalbegin<=i && i<=totalend) {
			return fr[i];
		} else {
			return 0;
		};
	};

	float signal_discrete::evaluate_imag(float x) const 
	{
		long i = long(x);
		if(totalbegin<=i && i<=totalend) {
			return fi[i];
		} else {
			return 0;
		};
	};

	void signal_discrete::write (FILE *file, const arguments &args, complex_component cc) const 
	{
/*		long i;
		long size  = abs(args.bmaxint - args.bminint);
		long begin = min(args.bmaxint, args.bminint);
		if (cc==cc_imag) {
			for (i=begin; i<size; i++) {
				fprintf(file, "%f\n", fi[i]);
			}
		} else {
			for (i=begin; i<size; i++) {
				fprintf(file, "%f\n", fr[i]);
			}
		}
		*/
	};

