// LyapunovKantz.cpp: Implementierungsdatei
//

#include "stdafx.h"
#include "Sounds.h"
#include "LyapunovKantz.h"

/*Author of Parts: Rainer Hegger. Last modified: Sep 3, 1999*/
#include <math.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define BOX 128
#define OUTSPEC 1

#define BOXSPEC 512
#define EPSMAXCONST 1.0

#define LEFTBORDER 40
#define BOTTOMBORDER 30
#define TOPBORDER 40

/* output is written every WHEN seconds */
#define WHEN 1
/* Size of the field for box assisted neighbour searching (has to be a power of 2)*/
#define NMAX 256
/* Size of the box for the scramble routine */
#define SCBOX 4096

/////////////////////////////////////////////////////////////////////////////
// CLyapunovKantz

CLyapunovKantz::CLyapunovKantz()
{
	m_intIBox=BOX-1;
	m_lLength=ULONG_MAX;
	m_lExclude=0;
	m_lReference=ULONG_MAX;
	m_intMaxdim=2;
	m_intMindim=2;
	m_intDelay=1;
	m_intColumn=1;
	m_intEpscount=5;
	m_intMaxiter=50;
	m_intWindow=0;
	m_intVerbosity=0xff;
	m_dEpsmin=0.001;
	m_dEpsmax=0.01;
	m_cEps0set=0;
	m_cEps1set=0;
	m_pOutfile=NULL;
	m_pInfile=NULL;

	//Lyapunovexponentenspektrumvoreinstellungen
	m_LSpec_epsset=0;
	m_LSpec_dimset=0;
	m_LSpec_COLUMNS=NULL;
	m_LSpec_LENGTH=ULONG_MAX;
	m_LSpec_exclude=0;
	m_LSpec_EMBED=3;
	m_LSpec_DIMENSION=1;
	m_LSpec_DELAY=2;
	m_LSpec_MINNEIGHBORS=30;
	m_LSpec_verbosity=0xff;
	m_LSpec_EPSSTEP=1.2;

	m_LSpec_avneig=0.0;
	m_LSpec_aveps=0.0;
	m_LSpec_imax=BOXSPEC-1;
	m_LSpec_count=0;

	//Korrelationsdimensionsberechnungsvoreinstellungen
	m_Korr_EMBED=10;
	m_Korr_HOWOFTEN=100;
	m_Korr_DELAY=1;
	m_Korr_EPSMAX=1.0;
	m_Korr_EPSMIN=0.001;
	m_Korr_MAXFOUND=1000;
	m_Korr_length=ULONG_MAX;
	m_Korr_DIM=1;				
	m_Korr_MINDIST=90;		
	m_Korr_imax=NMAX-1;

	//Sonstiges
	m_oldFaktorY=0;
	m_oldFaktorYKorr=0;
	m_KaplanYorkeDimension=0;
}

CLyapunovKantz::~CLyapunovKantz()
{
}

void CLyapunovKantz::SetParameters(LONG Iter,int MaxDim,int MinDim,int Tau,int NumberOf,double RadiusMin,
																	 double RadiusMax,INT Which,INT Referenz)
{
	switch(Which) 
	{
	case 0:
		//Lyapunovexponentenspektrumvoreinstellungen
		m_LSpec_epsset=0;
		m_LSpec_dimset=0;
		m_LSpec_COLUMNS=NULL;
		m_LSpec_LENGTH=Iter;
		m_LSpec_exclude=0;
		m_LSpec_EMBED=MinDim;
		m_LSpec_DIMENSION=1;
		m_LSpec_DELAY=Tau;
		m_LSpec_MINNEIGHBORS=NumberOf;
		m_LSpec_verbosity=0xff;
		m_LSpec_EPSSTEP=1.2;
		m_LSpec_epsmin=RadiusMin;
		m_LSpec_avneig=0.0;
		m_LSpec_aveps=0.0;
		m_LSpec_imax=BOXSPEC-1;
		m_LSpec_count=0;
		break;
	case 1:
		m_intIBox=BOX-1;
		m_lLength=ULONG_MAX;
		m_lExclude=0;
//		m_lReference=ULONG_MAX;
		m_lReference=Referenz;
		m_intMaxdim=MaxDim;
		m_intMindim=MinDim;
		m_intDelay=Tau;
		m_intColumn=1;
		m_intEpscount=NumberOf;
//		m_intMaxiter=Iter;
		m_intMaxiter=Iter;
		m_intWindow=0;
		m_intVerbosity=0xff;
		m_dEpsmin=RadiusMin;
		m_dEpsmax=RadiusMax;
		m_cEps0set=0;
		m_cEps1set=0;
		m_pOutfile=NULL;
		m_pInfile=NULL;
		break;
	case 2:
		//Korrelationsdimensionsberechnungsvoreinstellungen
		m_Korr_EMBED=MaxDim;					//Maximale Einbettungsdimension
		m_Korr_HOWOFTEN=NumberOf;			//Anzahl der Radien zwischen Rmin und RMax
		m_Korr_DELAY=Tau;							//Delayzeit
		m_Korr_EPSMAX=RadiusMax;			//Maximaler Suchradius
		m_Korr_EPSMIN=RadiusMin;			//Minimaler Suchradius
		m_Korr_MAXFOUND=Iter;					//Maximale Anzahl der untersuchten Paare
		m_Korr_DIM=1;									//bei Multivariate Daten >1
		m_Korr_MINDIST=500;							//Theiler Fenstergre
		m_Korr_imax=NMAX-1;
		break;
	}



	
	//Sonstiges
	m_KaplanYorkeDimension=0;
	m_oldFaktorY=0;
	m_oldFaktorYKorr=0;
}

/////////////////////////////////////////////////////////////////////////////
// 
void CLyapunovKantz::check_alloc(void *pnt)
{
  if (pnt == NULL) 
	{
    AfxMessageBox(E_LYAPUNOVKANTZ_CHECKALLOC);
		g_bIsBreak=TRUE;
	}
}

void CLyapunovKantz::put_in_boxes(double eps)
{
  unsigned long i;
  long j,k;
  static unsigned long blength;

  blength=m_lLength-(m_intMaxdim-1)*m_intDelay-m_intMaxiter;

  for (i=0;i<BOX;i++)
    for (j=0;j<BOX;j++)
      m_lBox[i][j]= -1;

  for (i=0;i<blength;i++) {
    j=(long)(m_pdSeries[i]/eps)&m_intIBox;
    k=(long)(m_pdSeries[i+m_intDelay]/eps)&m_intIBox;
    m_plListe[i]=m_lBox[j][k];
    m_lBox[j][k]=i;
  }
}
 
void CLyapunovKantz::lfind_neighbors(long act,double eps)
{
  unsigned int hi,k,k1;
  long i,j,i1,i2,j1,element;
  static long lwindow;
  double dx;
	double eps2=pow(eps,2);

  lwindow=(long)m_intWindow;
  for (hi=0;hi<m_intMaxdim-1;hi++)
    m_plFound[hi]=0;
  i=(long)(m_pdSeries[act]/eps)&m_intIBox;
  j=(long)(m_pdSeries[act+m_intDelay]/eps)&m_intIBox;
  for (i1=i-1;i1<=i+1;i1++) 
	{
    i2=i1&m_intIBox;
    for (j1=j-1;j1<=j+1;j1++) 
		{
      element=m_lBox[i2][j1&m_intIBox];
      while (element != -1) 
			{
				if ((element < (act-lwindow)) || (element > (act+lwindow))) 
				{
					dx=pow(m_pdSeries[act]-m_pdSeries[element],2);
					if (dx <= eps2) 
					{
						for (k=1;k<m_intMaxdim;k++) 
						{
							k1=k*m_intDelay;
							dx += pow(m_pdSeries[act+k1]-m_pdSeries[element+k1],2);
							if (dx <= eps2) 
							{
								k1=k-1;
								m_pplLFound[k1][m_plFound[k1]]=element;
								m_plFound[k1]++;
							}
							else
								break;
						}
					}
				}
				element=m_plListe[element];
      }
    }
  }
}

void CLyapunovKantz::iterate_points(long act)
{
  double **lfactor;
  double *dx;
  unsigned int i,j,l,l1;
  long k,element,**lcount;
  
  check_alloc(lfactor=(double**)malloc(sizeof(double*)*(m_intMaxdim-1)));
  check_alloc(lcount=(long**)malloc(sizeof(long*)*(m_intMaxdim-1)));
  for (i=0;i<m_intMaxdim-1;i++) 
	{
    check_alloc(lfactor[i]=(double*)malloc(sizeof(double)*(m_intMaxiter+1)));
    check_alloc(lcount[i]=(long*)malloc(sizeof(long)*(m_intMaxiter+1)));
  }
  check_alloc(dx=(double*)malloc(sizeof(double)*(m_intMaxiter+1)));

  for (i=0;i<=m_intMaxiter;i++)
    for (j=0;j<m_intMaxdim-1;j++) 
		{
      lfactor[j][i]=0.0;
      lcount[j][i]=0;
    }
  
  for (j=m_intMindim-2;j<m_intMaxdim-1;j++) 
	{
    for (k=0;k<m_plFound[j];k++) 
		{
      element=m_pplLFound[j][k];
      for (i=0;i<=m_intMaxiter;i++)
				dx[i]=pow(m_pdSeries[act+i]-m_pdSeries[element+i],2);
      for (l=1;l<j+2;l++) 
			{
				l1=l*m_intDelay;
				for (i=0;i<=m_intMaxiter;i++)
					dx[i] += pow(m_pdSeries[act+i+l1]-m_pdSeries[element+l1+i],2);
      }
      for (i=0;i<=m_intMaxiter;i++)
				if (dx[i] > 0.0)
				{
				  lcount[j][i]++;
					lfactor[j][i] += dx[i];
				}
		}
  }

  for (i=m_intMindim-2;i<m_intMaxdim-1;i++)
    for (j=0;j<=m_intMaxiter;j++)
      if (lcount[i][j]) 
			{
				m_pplCount[i][j]++;
				m_ppdLyap[i][j] += log(lfactor[i][j]/lcount[i][j])/2.0;
      }
  
  for (i=0;i<m_intMaxdim-1;i++)
	{
    free(lfactor[i]);
    free(lcount[i]);
  }
  free(lcount);
  free(lfactor);
  free(dx);
}


void CLyapunovKantz::rescale_data(double *x,unsigned long l,double *min,double *interval)
{
  unsigned int i;
  
  *min=*interval=x[0];
  
  for (i=1;i<l;i++) {
    if (x[i] < *min) *min=x[i];
    if (x[i] > *interval) *interval=x[i];
  }
  *interval -= *min;

  if (*interval != 0.0) 
	{
    for (i=0;i<l;i++)
      x[i]=(x[i]- *min)/ *interval;
  }
  else 
	{
    AfxMessageBox(E_LYAPUNOVKANTZ_RESCALEDATA);	
		g_bIsBreak=TRUE;
  }
}

double** CLyapunovKantz::Calc_Max_Lyap_Exponent(CDC* pDC, HWND hwnd, double* series,unsigned long length,
																								int* pIterationen,int* Epsilons)
{
  char stdi=0;
  double eps_fak;
  double epsilon;
  unsigned int i,j,l;
	double** Outgoing;
	
//Parameter einlesen  scan_options(argc,argv);
	m_lLength=length;
	m_pdSeries=series;
//  series=get_series(infile,&length,exclude,column,verbosity);
  rescale_data(m_pdSeries,m_lLength,&m_dMin,&m_dMax);

  if (m_cEps0set)
    m_dEpsmin /= m_dMax;
  if (m_cEps1set)
    m_dEpsmax /= m_dMax;

  if (m_dEpsmin >= m_dEpsmax) 
	{
    m_dEpsmax=m_dEpsmin;
    m_intEpscount=1;
  }
  
  if (m_lReference > (m_lLength-m_intMaxiter-(m_intMaxdim-1)*m_intDelay))
    m_lReference=m_lLength-m_intMaxiter-(m_intMaxdim-1)*m_intDelay;
  if ((m_intMaxiter+(m_intMaxdim-1)*m_intDelay) >= m_lLength) 
	{
    AfxMessageBox(E_LYAPUNOVKANTZ_CALC_MAX_LAYP_EXPONENT);
		g_bIsBreak=TRUE;
  }

  if (m_intMaxdim < 2)
    m_intMaxdim=2;
  if (m_intMindim < 2)
    m_intMindim=2;
  if (m_intMindim > m_intMaxdim)
    m_intMaxdim=m_intMindim;
  
  check_alloc(m_plListe=(long*)malloc(sizeof(long)*(m_lLength)));
  check_alloc(m_plFound=(long*)malloc(sizeof(long)*(m_intMaxdim-1)));
  check_alloc(m_pplLFound=(long**)malloc(sizeof(long*)*(m_intMaxdim-1)));
  for (i=0;i<m_intMaxdim-1;i++)
    check_alloc(m_pplLFound[i]=(long*)malloc(sizeof(long)*(m_lLength)));
  check_alloc(m_pplCount=(long**)malloc(sizeof(long*)*(m_intMaxdim-1)));
  for (i=0;i<m_intMaxdim-1;i++)
    check_alloc(m_pplCount[i]=(long*)malloc(sizeof(long)*(m_intMaxiter+1)));
  check_alloc(m_ppdLyap=(double**)malloc(sizeof(double*)*(m_intMaxdim-1)));
  for (i=0;i<m_intMaxdim-1;i++)
    check_alloc(m_ppdLyap[i]=(double*)malloc(sizeof(double)*(m_intMaxiter+1)));

  check_alloc(Outgoing=(double**)malloc(sizeof(double*)*(m_intMaxdim-1)));
  for (i=0;i<m_intMaxdim-1;i++)
    check_alloc(Outgoing[i]=(double*)malloc(sizeof(double)*(m_intMaxiter+1)*(m_intEpscount+1)));

  if (m_intEpscount == 1)
    eps_fak=1.0;
  else
    eps_fak=pow(m_dEpsmax/m_dEpsmin,1.0/(double)(m_intEpscount-1));

//  FILE* fout=fopen("d:\\AdrianBauernhaus\\Testfiles\\Lyap_r.txt","w");
  for (l=0;l<m_intEpscount;l++) 
	{
		//eigentliche Berechnung
    epsilon=m_dEpsmin*pow(eps_fak,(double)l);
    for (i=0;i<m_intMaxdim-1;i++)
      for (j=0;j<=m_intMaxiter;j++) 
			{
				m_pplCount[i][j]=0;
				m_ppdLyap[i][j]=0.0;
      }
    put_in_boxes(epsilon);
    for (i=0;i<m_lReference;i++) 
		{
			//Abbruch mit ESC abfangen
			MSG msg;
			PeekMessage( &msg, NULL, 0, 0, PM_REMOVE );
	    if (msg.message == WM_KEYDOWN && msg.wParam == VK_ESCAPE) 
			{
				g_bIsBreak=TRUE;
				TranslateMessage(&msg); 
		    DispatchMessage(&msg); 
			}
			theApp.DoWaitCursor(0);
			if (g_bIsBreak) break;

      lfind_neighbors(i,epsilon);
			if (g_bIsBreak) break;
      iterate_points(i);
    }

		if (g_bIsBreak) break;

		for (i=m_intMindim-2;i<m_intMaxdim-1;i++) 
		{
//      fprintf(fout,"#epsilon= %e  dim= %d\n",epsilon*m_dMax,i+2);
      for (j=0;j<=m_intMaxiter;j++)
				if (m_pplCount[i][j])
				{
//					fprintf(fout,"%d %e %ld\n",j,m_ppdLyap[i][j]/m_pplCount[i][j],m_pplCount[i][j]);
					Outgoing[i][j+(l*(m_intMaxiter+1))]=m_ppdLyap[i][j]/m_pplCount[i][j];
				}
 //     fprintf(fout,"\n");
    }
 //   fflush(fout);
		DrawLyapunovMax(pDC,hwnd,Outgoing,m_intMaxiter,l+1,m_intMindim,m_intMaxdim);
  }
//  fclose(fout);

	free(m_plListe);
	free(m_plFound);
  for (i=0;i<m_intMaxdim-1;i++)
	{
    free(m_pplLFound[i]);
    free(m_pplCount[i]);
		free(m_ppdLyap[i]);
//		free(Outgoing[i]);
  }
//	free(Outgoing);
	free(m_pplLFound);
  free(m_pplCount);
  free(m_ppdLyap);

	*Epsilons=m_intEpscount;
	*pIterationen=m_intMaxiter;
	return Outgoing;
}

void CLyapunovKantz::DrawLyapunovMax(CDC* pDC,HWND hwnd, double** DATA,int Iterationen,int Epsilons,int MinDim,int MaxDim) 
{
	CRect rect;
	GetClientRect(hwnd,rect);

	//grten negativen Wert bestimmen
	double MaxData=0;
	m_intMindim=MinDim;
	m_intMaxdim=MaxDim;

  for (int i=(INT)(m_intMindim-2);i<(INT)(m_intMaxdim-1);i++) 
	{
    for (int j=0;j<=Iterationen*Epsilons;j++)
		{
			if (DATA[i][j]<MaxData) 
				MaxData=DATA[i][j];
		}
  }
	MaxData=fabs(MaxData);
	//Bildschirmskalierung berechnen
	double FaktorX=(rect.right-LEFTBORDER)/(Iterationen+1);
	double FaktorY=(rect.bottom-BOTTOMBORDER)/MaxData;

	CAutoFont font("Century Gothic");
	font.SetDC(pDC->GetSafeHdc());
	font.SetHeight(16);
	font.SetFontColor(RGB(0,0,0));
	font.SetEscapement(0);
	font.SetOrientation(0);
	pDC->SelectObject(&font);

	//Bildschirm lschen
	pDC->FillSolidRect(0,0,rect.right,rect.bottom,RGB(255,255,255));

	//berschrift
	CString Text;
	Text.LoadString(IDS_MAXLYAPUNOVEXPONENT);
 	pDC->DrawText(Text,rect,DT_TOP|DT_RIGHT);// TextOut(rect.right/2,16,Text);
	//Koordinatensystem zeichnen
	CPen RotStift,BlauStift,GruenStift,SchwarzStift,ExponentStift;
	ExponentStift.CreatePen(PS_DOT,1,RGB(0,0,0));
	RotStift.CreatePen(PS_SOLID,1,RGB(255,0,0));
	BlauStift.CreatePen(PS_SOLID,1,RGB(0,0,255));
	GruenStift.CreatePen(PS_SOLID,1,RGB(0,255,0));
	SchwarzStift.CreatePen(PS_SOLID,1,RGB(0,0,0));
	pDC->SelectObject(&SchwarzStift);
	font.SetEscapement(900);
	font.SetOrientation(900);
	pDC->SelectObject(&font);

	//Y-Achse
	pDC->MoveTo(LEFTBORDER,5);
	pDC->LineTo(LEFTBORDER,rect.bottom-BOTTOMBORDER+5);
	Text.LoadString(IDS_STRECHFAKTOR);
	pDC->TextOut(4,rect.bottom/2,Text);
	for(int j=1;j<(INT)(MaxData*4);j++)
	{
		if (j*FaktorY<rect.bottom-BOTTOMBORDER) 
		{
			pDC->MoveTo(LEFTBORDER,(LONG)(j*FaktorY));
			pDC->LineTo(LEFTBORDER-4,(LONG)(j*FaktorY));
			char buffer[20];
			CString Dimension;
			Dimension=itoa( j, buffer, 10 );
			Dimension="-"+Dimension;
			pDC->TextOut(LEFTBORDER-18,(LONG)(j*FaktorY),Dimension);
		}
	}
	//X-Achse
	font.SetEscapement(0);
	font.SetOrientation(0);
	pDC->SelectObject(&font);
	pDC->MoveTo(5,rect.bottom-BOTTOMBORDER);
	pDC->LineTo(rect.right-5,rect.bottom-BOTTOMBORDER);
	Text.LoadString(IDS_ITERATIONEN);
	pDC->TextOut(rect.right/2,rect.bottom-18,Text);
	for(j=0;j<=Iterationen;j+=10)
	{
		pDC->MoveTo((LONG)(LEFTBORDER+j*FaktorX),(LONG)(rect.bottom-BOTTOMBORDER));
		pDC->LineTo((LONG)(LEFTBORDER+j*FaktorX),(LONG)(rect.bottom-BOTTOMBORDER+4));
		char buffer[20];
		CString Dimension;
		Dimension=itoa( j, buffer, 10 );
		pDC->TextOut((LONG)(LEFTBORDER-3+j*FaktorX),(LONG)(rect.bottom-BOTTOMBORDER+2),Dimension);
	}

	//Kurven zeichnen
	for (int l=0;l<Epsilons;l++)
	{
		for (i=(INT)(m_intMindim-2);i<(INT)(m_intMaxdim-1);i++) 
		{
			if (i==(INT)(m_intMindim-2)) pDC->SelectObject(&RotStift);
			if (i==(INT)(m_intMindim-1)) pDC->SelectObject(&BlauStift);
			if (i>(INT)(m_intMindim-1)) pDC->SelectObject(&GruenStift);
			pDC->MoveTo(LEFTBORDER,rect.bottom-BOTTOMBORDER);
		  for (j=0;j<Iterationen;j++)
			{	
				pDC->LineTo((LONG)(LEFTBORDER+j*FaktorX),(LONG)(fabs(DATA[i][j+(l*Iterationen+1)])*FaktorY));
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//Lyapunov Spektrumermittlungsroutinen

double CLyapunovKantz::sort(long act,unsigned long nfound)
{
  double maxeps=0.0,dx,*abstand,dswap;
  unsigned int i,j,del,iswap,n1;
	int hf;

  check_alloc(abstand=(double*)malloc(nfound*sizeof(double)));

  for (i=0;i<nfound;i++) 
	{
    hf=m_LSpec_found[i];
    if (hf != act) 
		{
      abstand[i]=fabs(m_LSpec_Series[0][act]-m_LSpec_Series[0][hf]);
      for (j=1;j<m_LSpec_alldim;j++) 
			{
				del=m_LSpec_indexes[j][0];
				n1=m_LSpec_indexes[j][1];
				dx=fabs(m_LSpec_Series[n1][act-del]-m_LSpec_Series[n1][hf-del]);
				if (dx > abstand[i]) abstand[i]=dx;
      }
    }
    else
      abstand[i]=2.0*EPSMAXCONST;
  }
  for (i=0;i<nfound-1;i++)
    for (j=i+1;j<nfound;j++)
      if (abstand[j]<abstand[i]) 
			{
				dswap=abstand[i];
				abstand[i]=abstand[j];
				abstand[j]=dswap;
				iswap=m_LSpec_found[i];
				m_LSpec_found[i]=m_LSpec_found[j];
				m_LSpec_found[j]=iswap;
      }

  maxeps=abstand[m_LSpec_MINNEIGHBORS-1];
  free(abstand);

  return maxeps;
}

void CLyapunovKantz::make_multi_box(double **ser,long **box,long *list,unsigned long l,
										unsigned int bs,unsigned int dim,unsigned int emb,
										unsigned int del,double eps)
{
  unsigned int i,x,y;
  unsigned int ib=bs-1;

  for (x=0;x<bs;x++)
    for (y=0;y<bs;y++)
      box[x][y] = -1;
  
  for (i=(emb-1)*del;i<l;i++) {
    x=(int)(ser[0][i]/eps)&ib;
    y=(int)(ser[dim-1][i]/eps)&ib;
    list[i]=box[x][y];
    box[x][y]=i;
  }
}

unsigned long CLyapunovKantz::find_multi_neighbors(double **s,long **box,long *list,double **x,
																									 unsigned long l,unsigned int bs,unsigned int dim,
																									 unsigned int emb,unsigned int del,double eps,
																									 unsigned long *flist)
{
  unsigned long nf=0;
  int i,i1,i2,j,j1,k,k1,li;
  int ib=bs-1;
  long element;
  double dx;
  
  i=(int)(x[0][0]/eps)&ib;
  j=(int)(x[dim-1][0]/eps)&ib;
  
  for (i1=i-1;i1<=i+1;i1++) 
	{
    i2=i1&ib;
    for (j1=j-1;j1<=j+1;j1++) 
		{
      element=box[i2][j1&ib];
      while (element != -1) 
			{
				for (k=0;k<(INT)emb;k++) 
				{
					k1= -k*(int)del;
					for (li=0;li<(INT)dim;li++) 
					{
						dx=fabs(x[li][k1]-s[li][element+k1]);
						if (dx > eps)
							break;
					}
					if (dx > eps)
						break;
				}
				if (dx <= eps)
				  flist[nf++]=element;
				element=list[element];
      }
    }
  }
  return nf;
}

void CLyapunovKantz::solvele(double **mat5,double *vec2,unsigned int n)
{
  double vswap,*mswap,*hvec,max,h,pivot,q;
  int i,j,k,maxi;

  for (i=0;i<(INT)(n-1);i++) 
	{
    max=fabs(mat5[i][i]);
    maxi=i;
    for (j=i+1;j<(INT)n;j++)
      if ((h=fabs(mat5[j][i])) > max) 
			{
				max=h;
				maxi=j;
      }
    if (maxi != i) 
		{
      mswap=mat5[i];
      mat5[i]=mat5[maxi];
      mat5[maxi]=mswap;
      vswap=vec2[i];
      vec2[i]=vec2[maxi];
      vec2[maxi]=vswap;
    }
    
    hvec=mat5[i];
    pivot=hvec[i];
    if (fabs(pivot) == 0.0) 
		{
      AfxMessageBox(E_LYAPUNOVKANTZ_SOLVELE);
			g_bIsBreak=TRUE;
			break;
    }
//    for (int z=0;z<=(INT)m_LSpec_alldim;z++)
//		  _RPTF4(_CRT_WARN, "mat5 vorher: '%g,%g,%g,%g'\n",  mat5[z][0],mat5[z][1],mat5[z][2],mat5[z][3]);
    for (j=i+1;j<(INT)n;j++) 
		{
      q= -mat5[j][i]/pivot;
//			if (q==0) 
//				BOOL NoNo=TRUE;
      mat5[j][i]=0.0;
      for (k=i+1;k<(INT)n;k++)
				mat5[j][k] += q*hvec[k];
      vec2[j] += q*vec2[i];
    }
//    for (z=0;z<=(INT)m_LSpec_alldim;z++)
//		  _RPTF4(_CRT_WARN, "mat5 nachher: '%g,%g,%g,%g'\n",  mat5[z][0],mat5[z][1],mat5[z][2],mat5[z][3]);
  }
	if (mat5[n-1][n-1]==0)								//hinzugefgt
		mat5[n-1][n-1]=0.000000000001;

  vec2[n-1] /= mat5[n-1][n-1];
	

  for (i=n-2;i>=0;i--) 
	{
    hvec=mat5[i];
    for (j=n-1;j>i;j--)
      vec2[i] -= hvec[j]*vec2[j];
    vec2[i] /= hvec[i];
//		if (_finite(vec2[i])==0) 
//				BOOL NoNo=TRUE;
  }
}

double** CLyapunovKantz::invert_matrix(double **mat2,unsigned int size)
{
  int i,j,k;
  double **hmat,**imat,*vec2;

  check_alloc(hmat=(double**)malloc(sizeof(double*)*size));
  for (i=0;i<(INT)size;i++) 
	{
    check_alloc(hmat[i]=(double*)malloc(sizeof(double)*size));
  }

  check_alloc(imat=(double**)malloc(sizeof(double*)*size));
  for (i=0;i<(INT)size;i++) 
	{
    check_alloc(imat[i]=(double*)malloc(sizeof(double)*size));
  }

  check_alloc(vec2=(double*)malloc(sizeof(double)*size));
  
  for (i=0;i<(INT)size;i++) 
	{
    for (j=0;j<(INT)size;j++) 
		{
      vec2[j]=(i==j)?1.0:0.0;
      for (k=0;k<(INT)size;k++)
				hmat[j][k]=mat2[j][k];
    }
//    for (int z=0;z<=(INT)m_LSpec_alldim;z++)
//		{
//		  _RPTF4(_CRT_WARN, "hmat vor solvele: '%g,%g,%g,%g'\n",  hmat[z][0],hmat[z][1],hmat[z][2],hmat[z][3]);
//		}
    solvele(hmat,vec2,size);
//    for (z=0;z<=(INT)m_LSpec_alldim;z++)
//		{
//		  _RPTF4(_CRT_WARN, "hmat nach solvele: '%g,%g,%g,%g'\n",  hmat[z][0],hmat[z][1],hmat[z][2],hmat[z][3]);
//		}

   for (j=0;j<(INT)size;j++)
		{
      imat[j][i]=vec2[j];
//			if (_finite(imat[j][i])==0) 
//				BOOL NoNo=TRUE;
		}
	 if (g_bIsBreak) break;
  }
  
  free(vec2);
  for (i=0;i<(INT)size;i++)
    free(hmat[i]);
  free(hmat);

  return imat;
}

void CLyapunovKantz::make_dynamics(double **dynamics,long act)
{
  long i,hi,j,hj,k,t=act,d;
  unsigned long nfound=0;
  double **hser,**imat;
  double epsilon,hv,hv1;
  double neu;
	double*	vec;
	double** mat;
	
  check_alloc(vec=(double*)malloc(sizeof(double)*(m_LSpec_alldim+1)));  

  check_alloc(mat=(double**)malloc(sizeof(double*)*(m_LSpec_alldim+1)));
	for (i=0;i<=(INT)m_LSpec_alldim;i++)
    check_alloc(mat[i]=(double*)malloc(sizeof(double)*(m_LSpec_alldim+1)));

  check_alloc(hser=(double**)malloc(sizeof(double*)*m_LSpec_DIMENSION));
  for (i=0;i<(INT)m_LSpec_DIMENSION;i++)
    hser[i]=m_LSpec_Series[i]+act;
  for (epsilon=m_LSpec_epsmin;epsilon<=EPSMAXCONST;epsilon*=m_LSpec_EPSSTEP) 
	{
    make_multi_box(m_LSpec_Series,m_LSpec_box,m_LSpec_list,m_LSpec_LENGTH-m_LSpec_DELAY,BOXSPEC,
										m_LSpec_DIMENSION,m_LSpec_EMBED,m_LSpec_DELAY,epsilon);
    nfound=find_multi_neighbors(m_LSpec_Series,m_LSpec_box,m_LSpec_list,hser,m_LSpec_LENGTH-m_LSpec_DELAY,
																BOXSPEC,m_LSpec_DIMENSION,m_LSpec_EMBED,m_LSpec_DELAY,epsilon,m_LSpec_found);
    if (nfound > m_LSpec_MINNEIGHBORS) 
		{
      epsilon=sort(act,nfound);
      nfound=m_LSpec_MINNEIGHBORS;
      break;
    }
  }
  free(hser);

  m_LSpec_avneig += nfound;
  m_LSpec_aveps += epsilon;
  m_LSpec_epsmin=m_LSpec_aveps/m_LSpec_count;
  if (nfound < m_LSpec_MINNEIGHBORS) 
	{
    AfxMessageBox(E_LYAPUNOVKANTZ_MAKEDYNAMICS);
		g_bIsBreak=TRUE;
  }
  
  for (i=0;i<=(INT)m_LSpec_alldim;i++) 
	{
    vec[i]=0.0;
    for (j=0;j<=(INT)m_LSpec_alldim;j++) 
      mat[i][j]=0.0;
  }
  
  for (i=0;i<(INT)nfound;i++) 
	{
    act=m_LSpec_found[i];
    mat[0][0] += 1.0;
    for (j=0;j<(INT)m_LSpec_alldim;j++)
      mat[0][j+1] += m_LSpec_Series[m_LSpec_indexes[j][1]][act-m_LSpec_indexes[j][0]];
    for (j=0;j<(INT)m_LSpec_alldim;j++) 
		{
      hv1=m_LSpec_Series[m_LSpec_indexes[j][1]][act-m_LSpec_indexes[j][0]];
      hj=j+1;
      for (k=j;k<(INT)m_LSpec_alldim;k++)
				mat[hj][k+1] += m_LSpec_Series[m_LSpec_indexes[k][1]][act-m_LSpec_indexes[k][0]]*hv1;
    }
  }

  for (i=0;i<=(INT)m_LSpec_alldim;i++)
    for (j=i;j<=(INT)m_LSpec_alldim;j++)
		{
      mat[j][i]=(mat[i][j]/=(double)nfound);
//			if (mat[3][3]==0) 
//				BOOL NoNo=TRUE;
//			if (_finite(mat[j][i])==0) 
//				BOOL NoNo=TRUE;
//			if (nfound==0) 
//				BOOL NoNo=TRUE;
		}
//  for (i=0;i<=(INT)m_LSpec_alldim;i++)
//	{
////		  _RPTF4(_CRT_WARN, "mat vor Invert: '%g,%g,%g,%g'\n",  mat[i][0],mat[i][1],mat[i][2],mat[i][3]);
//	}
	imat=invert_matrix(mat,m_LSpec_alldim+1);
//  for (i=0;i<=(INT)m_LSpec_alldim;i++)
//	{
////		  _RPTF4(_CRT_WARN, "imat nach Invert: '%g,%g,%g,%g'\n",  imat[i][0],imat[i][1],imat[i][2],imat[i][3]);
//	}
//
  for (d=0;d<(INT)m_LSpec_DIMENSION;d++) 
	{
    for (i=0;i<=(INT)m_LSpec_alldim;i++)
      vec[i]=0.0;
    for (i=0;i<(INT)nfound;i++) 
		{
      act=m_LSpec_found[i];
      hv=m_LSpec_Series[d][act+m_LSpec_DELAY];
      vec[0] += hv;
      for (j=0;j<(INT)m_LSpec_alldim;j++)
				vec[j+1] += hv*m_LSpec_Series[m_LSpec_indexes[j][1]][act-m_LSpec_indexes[j][0]];
    }
    for (i=0;i<=(INT)m_LSpec_alldim;i++)
      vec[i] /= (double)nfound;
    
    neu=0.0;
    for (i=0;i<=(INT)m_LSpec_alldim;i++)
      neu += imat[0][i]*vec[i];
    for (i=1;i<=(INT)m_LSpec_alldim;i++) 
		{
      hi=i-1;
      dynamics[d][hi]=0.0;
      for (j=0;j<=(INT)m_LSpec_alldim;j++)
			{
				dynamics[d][hi] += imat[i][j]*vec[j];
//				if (_finite(dynamics[d][hi])==0)
//					BOOL NoNo=TRUE;
			}
    }
    for (i=0;i<(INT)m_LSpec_alldim;i++)
      neu += dynamics[d][i]*m_LSpec_Series[m_LSpec_indexes[i][1]][t-m_LSpec_indexes[i][0]];
    m_LSpec_averr[d] += (neu-m_LSpec_Series[d][t+m_LSpec_DELAY])*(neu-m_LSpec_Series[d][t+m_LSpec_DELAY]);
  }

  for (i=0;i<=(INT)m_LSpec_alldim;i++)
    free(imat[i]);
  free(imat);

  for (i=0;i<=(INT)m_LSpec_alldim;i++)
    free(mat[i]);
  free(mat);

	free(vec);
}


void CLyapunovKantz::gram_schmidt(double **delta,double *stretch)
{
  double **dnew,norm,*diff;
  long i,j,k;
  
  check_alloc(diff=(double*)malloc(sizeof(double)*m_LSpec_alldim));
  check_alloc(dnew=(double**)malloc(sizeof(double*)*m_LSpec_alldim));
  for (i=0;i<(INT)m_LSpec_alldim;i++)
    check_alloc(dnew[i]=(double*)malloc(sizeof(double)*m_LSpec_alldim));

//  for (i=0;i<(INT)m_LSpec_alldim;i++)  //hinzugefgt
//    for (j=0;j<(INT)m_LSpec_alldim;j++)//hinzugefgt
//      dnew[i][j]=0.0;					//hinzugefgt
	for (i=0;i<(INT)m_LSpec_alldim;i++) 
	{
    for (j=0;j<(INT)m_LSpec_alldim;j++) 
      diff[j]=0.0;
    for (j=0;j<i;j++) 
		{
      norm=0.0;
      for (k=0;k<(INT)m_LSpec_alldim;k++)
			{
				norm += delta[i][k]*dnew[j][k];
//				if (_finite(delta[i][k])==0) 
//					BOOL NoNo=TRUE;
			}
      for (k=0;k<(INT)m_LSpec_alldim;k++)
				diff[k] -= norm*dnew[j][k];
    }
    norm=0.0;
    for (j=0;j<(INT)m_LSpec_alldim;j++)
		{
			double Test=delta[i][j]+diff[j];
			double Test2=pow(delta[i][j]+diff[j],2);
      norm += pow(delta[i][j]+diff[j],2);
//			if (j==2&&norm==0) 
//				BOOL NoNo=TRUE;
		}
    stretch[i]=(norm=sqrt(norm+0.000000000000000001));
//    stretch[i]=(_finite(sqrt(norm))==0?sqrt(norm):0);  //modifiziert
//		norm=_finite(sqrt(norm))==0?sqrt(norm):1;					//modifiziert
//		if (stretch[i]<0)
//			BOOL NoNo=TRUE;
    for (j=0;j<(INT)m_LSpec_alldim;j++)
		{
      dnew[i][j]=(delta[i][j]+diff[j])/norm;
//			if (_finite(dnew[i][j])==0)
//				BOOL NoNo=TRUE;
		}
  }
  for (i=0;i<(INT)m_LSpec_alldim;i++)
    for (j=0;j<(INT)m_LSpec_alldim;j++)
      delta[i][j]=dnew[i][j];

  free(diff);
  for (i=0;i<(INT)m_LSpec_alldim;i++)
    free(dnew[i]);
  free(dnew);
}


void CLyapunovKantz::make_iteration(double **dynamics,double **delta)
{
  double **dnew;
  long i,j,k;

  check_alloc(dnew=(double**)malloc(sizeof(double*)*m_LSpec_alldim));
  for (i=0;i<(INT)m_LSpec_alldim;i++)
    check_alloc(dnew[i]=(double*)malloc(sizeof(double)*m_LSpec_alldim));

  for (i=0;i<(INT)m_LSpec_alldim;i++) 
	{
    for (j=0;j<(INT)m_LSpec_DIMENSION;j++) 
		{
      dnew[i][j]=dynamics[j][0]*delta[i][0];
//			if (_finite(delta[i][0])==0) 
//				BOOL NoNo=TRUE;
//
      for (k=1;k<(INT)m_LSpec_alldim;k++)
			{
				dnew[i][j] += dynamics[j][k]*delta[i][k];
//				if (_finite(delta[i][k])==0) 
//					BOOL NoNo=TRUE;
			}
    }
    for (j=m_LSpec_DIMENSION;j<(INT)m_LSpec_alldim;j++)
      dnew[i][j]=delta[i][j-1];
  }

  for (i=0;i<(INT)m_LSpec_alldim;i++)
    for (j=0;j<(INT)m_LSpec_alldim;j++)
		{
      delta[i][j]=dnew[i][j];
//			if (_finite(delta[i][j])==0) 
//				BOOL NoNo=TRUE;
		}

  for (i=0;i<(INT)m_LSpec_alldim;i++)
    free(dnew[i]);
  free(dnew);
}

void CLyapunovKantz::variance(double *s,unsigned long l,double *av,double *var)
{
  unsigned long i;
  double h;
  
  *av= *var=0.0;

  for (i=0;i<l;i++) 
	{
    h=s[i];
    *av += h;
    *var += h*h;
  }
  *av /= (double)l;
  *var=sqrt(fabs((*var)/(double)l-(*av)*(*av)));
  if (*var == 0.0) 
	{
    AfxMessageBox(E_LYAPUNOVKANTZ_VARIANCE);
		g_bIsBreak=TRUE;
  }
}

#ifndef M_PI
#define M_PI 3.1415926535897932385E0
#endif


void CLyapunovKantz::rnd_init(unsigned long iseed)
{
  int i;
  unsigned long z,index;

  if (sizeof(long) == 8) 
	{
    factor=13*13*13*13;
    factor=factor*factor*factor*13;
  }
  else
    factor=69069;
  lo_limit=(double)ULONG_MAX;

  check_alloc(rnd_array=(unsigned long *)malloc(9689*sizeof(unsigned long)));
  check_alloc(nexti=(unsigned long *)malloc(9689*sizeof(long)));
  check_alloc(rnd1279=(unsigned long *)malloc(1279*sizeof(unsigned long)));
  check_alloc(next1279=(unsigned long *)malloc(1279*sizeof(long)));

  rnd_array[0]=rnd1279[0]=iseed;
  rnd69=iseed;
  index=iseed;
  nexti[0]=next1279[0]=1;

  for (i=1;i<9689;i++) 
	{
    rnd_array[i]=factor*rnd_array[i-1]+1;
    nexti[i]=i+1;
  }

  for (i=1;i<1279;i++) 
	{
    rnd1279[i]=factor*rnd1279[i-1]+1;
    next1279[i]=i+1;
  }
  nexti[9688]=next1279[1278]=0;

  for (i=1;i<2000;i++) 
	{
    index=factor*index+1;
    z=rnd1279[((index>>10)%1279)];
    z=(z<<10)+(z>>10);
    index=factor*index+1;
    rnd1279[((index>>10)%1279)] += z;
  }

  nexti[9688]=next1279[1278]=0;
  rndtime=9688;
  rndtime1=9688-157;
  rndtime2=9688-314;
  rndtime3=9688-471;
  t1279=1278;
  t1279_1=1278-216;
  t1279_2=1278-299;
  t1279_3=1278-598;
}

unsigned long CLyapunovKantz::rnd_long(void)
{
  rndtime=nexti[rndtime];
  rndtime1=nexti[rndtime1];
  rndtime2=nexti[rndtime2];
  rndtime3=nexti[rndtime3];
  rnd_array[rndtime] ^= rnd_array[rndtime1]
    ^rnd_array[rndtime2]^rnd_array[rndtime3];

  return rnd_array[rndtime];
}

double** CLyapunovKantz::Calc_Lyap_Spektrum(CDC* pDC, HWND hwnd, double** Series, int iters, int Laenge)
{
  char stdi=0;
  double **delta,**dynamics,*lfactor;
  double *factor,dim;
  double *hseries;
  double *interval,*min,*av,*var,maxinterval;
  long start,i,j;
  time_t lasttime;//,newtime;
//	FILE* file=fopen("d:\\AdrianBauernhaus\\Testfiles\\LyapunovSpektrum.txt","w");
	CRect Fenster;

  m_LSpec_ITERATIONS=ULONG_MAX;
  m_LSpec_alldim=m_LSpec_DIMENSION*m_LSpec_EMBED;

//  if (COLUMNS == NULL)
//    SpectrumSeries=(double**)get_multi_series("d:\\AdrianBauernhaus\\Zeitreihen\\LORENZ.DAT",
//																							&LENGTH,exclude,&DIMENSION,"",dimset,verbosity);
//  else
//    SpectrumSeries=(double**)get_multi_series(infile,&LENGTH,exclude,&DIMENSION,
//				      COLUMNS,dimset,verbosity);
  m_LSpec_Series=Series;
	m_LSpec_LENGTH=Laenge;
  if (m_LSpec_MINNEIGHBORS > (m_LSpec_LENGTH-m_LSpec_DELAY*(m_LSpec_EMBED-1)-1)) 
	{
    AfxMessageBox(E_LYAPUNOVKANTZ_CALCLYAPSPEKTRUM);
		g_bIsBreak=TRUE;
  }

  check_alloc(min=(double*)malloc(sizeof(double)*m_LSpec_DIMENSION));
  check_alloc(interval=(double*)malloc(sizeof(double)*m_LSpec_DIMENSION));
  check_alloc(av=(double*)malloc(sizeof(double)*m_LSpec_DIMENSION));
  check_alloc(var=(double*)malloc(sizeof(double)*m_LSpec_DIMENSION));
  check_alloc(m_LSpec_averr=(double*)malloc(sizeof(double)*m_LSpec_DIMENSION));
  maxinterval=0.0;
  for (i=0;i<(INT)m_LSpec_DIMENSION;i++) 
	{
    m_LSpec_averr[i]=0.0;
    rescale_data(m_LSpec_Series[i],m_LSpec_LENGTH,&min[i],&interval[i]);
    if (interval[i] > maxinterval) 
      maxinterval=interval[i];
    variance(m_LSpec_Series[i],m_LSpec_LENGTH,&av[i],&var[i]);
  }

  
  if (m_LSpec_INVERSE) 
	{
    check_alloc(hseries=(double*)malloc(sizeof(double)*m_LSpec_LENGTH));
    for (j=0;j<(INT)m_LSpec_DIMENSION;j++) 
		{
      for (i=0;i<(INT)m_LSpec_LENGTH;i++)
				hseries[m_LSpec_LENGTH-1-i]=m_LSpec_Series[j][i];
      for (i=0;i<(INT)m_LSpec_LENGTH;i++)
				m_LSpec_Series[j][i]=hseries[i];
    }
    free(hseries);
  }
  
  if (!m_LSpec_epsset)
    m_LSpec_epsmin=1./1000.;
  else
    m_LSpec_epsmin /= maxinterval;
  
  check_alloc(m_LSpec_box=(long**)malloc(sizeof(long*)*BOXSPEC));
  for (i=0;i<BOXSPEC;i++)
    check_alloc(m_LSpec_box[i]=(long*)malloc(sizeof(long)*BOXSPEC));

  check_alloc(m_LSpec_list=(long*)malloc(sizeof(long)*m_LSpec_LENGTH));
  check_alloc(m_LSpec_found=(unsigned long*)malloc(sizeof(long)*m_LSpec_LENGTH));

  check_alloc(dynamics=(double**)malloc(sizeof(double*)*m_LSpec_DIMENSION));
  for (i=0;i<(INT)m_LSpec_DIMENSION;i++)
    check_alloc(dynamics[i]=(double*)malloc(sizeof(double)*m_LSpec_alldim));
  check_alloc(factor=(double*)malloc(sizeof(double)*m_LSpec_alldim));
  check_alloc(lfactor=(double*)malloc(sizeof(double)*m_LSpec_alldim));
  check_alloc(delta=(double**)malloc(sizeof(double*)*m_LSpec_alldim));
  for (i=0;i<(INT)m_LSpec_alldim;i++)
    check_alloc(delta[i]=(double*)malloc(sizeof(double)*m_LSpec_alldim));
  
  
  check_alloc(m_LSpec_indexes=(unsigned int**)malloc(sizeof(int*)*m_LSpec_alldim));
  for (i=0;i<(INT)m_LSpec_alldim;i++) 
	{
    check_alloc(m_LSpec_indexes[i]=(unsigned int*)malloc(sizeof(int)*2));
    m_LSpec_indexes[i][0]=(unsigned int)(i/m_LSpec_DIMENSION)*m_LSpec_DELAY;
    m_LSpec_indexes[i][1]=i%m_LSpec_DIMENSION;
  }

  rnd_init(0x098342L);
  for (i=0;i<10000;i++)
    rnd_long();
  for (i=0;i<(INT)m_LSpec_alldim;i++) 
	{
    factor[i]=0.0;
    for (j=0;j<(INT)m_LSpec_alldim;j++)
		{
	    delta[i][j]=(double)rnd_long()/(double)ULONG_MAX;
//			if (_finite(delta[i][j])==0) 
//				BOOL NoNo=TRUE;
		}
		//if (i<=j) delta[i][j]=1.0;
      //else delta[i][j]=0.0;
  }
  gram_schmidt(delta,lfactor);
  m_LSpec_ITERATIONS=iters;
  start=m_LSpec_ITERATIONS;
  if (start>(INT)(m_LSpec_LENGTH-m_LSpec_DELAY)) 
    start=m_LSpec_LENGTH-m_LSpec_DELAY;


	double** LYAEXP;
	LYAEXP=(double**)malloc(sizeof(double*)*m_LSpec_ITERATIONS);
	for (i=0;i<(INT)m_LSpec_ITERATIONS;i++)
		LYAEXP[i]=(double*)malloc(sizeof(double)*m_LSpec_alldim);


  m_LSpec_count=0;

  time(&lasttime);
  for (i=(m_LSpec_EMBED-1)*m_LSpec_DELAY;i<start;i++) 
	{
    m_LSpec_count++;
    make_dynamics(dynamics,i);
		if (g_bIsBreak) break;

//		_RPTF3(_CRT_WARN, "dynamics: '%g,%g,%g'\n", dynamics[0][0],dynamics[0][1],dynamics[0][2]);
//		for (int z=0;z<(INT)m_LSpec_alldim;z++) {
//		}
//		  _RPTF4(_CRT_WARN, "delta:%u '%g,%g,%g'\n",i, delta[z][0],delta[z][1],delta[z][2]);
    
		make_iteration(dynamics,delta);
		if (g_bIsBreak) break;
	  
//		for (z=0;z<(INT)m_LSpec_alldim;z++) {
//		}
//			  _RPTF4(_CRT_WARN, "delta2:%u '%g,%g,%g'\n",i, delta[z][0],delta[z][1],delta[z][2]);
    
		gram_schmidt(delta,lfactor);
		if (g_bIsBreak) break;

    for (j=0;j<(INT)m_LSpec_alldim;j++) 
		{
      factor[j] += log(lfactor[j])/(double)m_LSpec_DELAY;
			LYAEXP[m_LSpec_count-1][j]=factor[j]/m_LSpec_count;									//  !!!!!!!!!!
    }
	
		//Abbruch mit ESC abfangen
		MSG msg;
		PeekMessage( &msg, NULL, 0, 0, PM_REMOVE );
    if (msg.message == WM_KEYDOWN && msg.wParam == VK_ESCAPE) 
		{
			g_bIsBreak=TRUE;
	    TranslateMessage(&msg); 
	    DispatchMessage(&msg); 
		}
		theApp.DoWaitCursor(0);
		if (g_bIsBreak) break;

		if (m_LSpec_count>10) 
			DrawExponents(pDC,hwnd,LYAEXP,m_LSpec_count,start,m_LSpec_alldim,m_LSpec_DELAY);
  }
  
	if (!g_bIsBreak)
	{
		dim=0.0;
		for (i=0;i<(INT)m_LSpec_alldim;i++) 
		{
		  dim += factor[i];
			if (dim < 0.0)
				break;
	  }
		if (i < (INT)m_LSpec_alldim)
			dim=i+(dim-factor[i])/fabs(factor[i]);
		 else
		  dim=m_LSpec_alldim;
/*

  fprintf(file,"#Average relative forecast errors:= ");
  for (i=0;i<(INT)DIMENSION;i++)
    fprintf(file,"%e ",sqrt(averr[i]/count)/var[i]);
*/
		DOUBLE Nachbarschaft=m_LSpec_aveps*maxinterval/(double)m_LSpec_count;
		m_KaplanYorkeDimension=dim;

		CRect rect;
		GetClientRect(hwnd,rect);
		int dec;
		int sign;
		CString strWert=ecvt(m_KaplanYorkeDimension,4, &dec,&sign );
		if (dec<=0) 
			for(int z=0;z<=abs(dec);z++) strWert.Insert(0,'0');
		if (dec<=0) strWert.Insert(1,'.');
		else strWert.Insert(dec,'.');
		if (sign!=0) strWert.Insert(0,'-');
		CString strNachbar=ecvt(Nachbarschaft,4, &dec,&sign );
		if (dec<=0) 
			for(int z=0;z<=abs(dec);z++) strNachbar.Insert(0,'0');
		if (dec<=0) strNachbar.Insert(1,'.');
		else strNachbar.Insert(dec,'.');
		if (sign!=0) strNachbar.Insert(0,'-');
		CString Text;
		AfxFormatString2(Text,IDS_NEAREST_NEIGHBORS,strWert,strNachbar);
		pDC->DrawText(Text,rect,DT_TOP|DT_CENTER);
	}

  for (i=0;i<(INT)m_LSpec_alldim;i++) 
    free(m_LSpec_indexes[i]);
  free(m_LSpec_indexes);

  for (i=0;i<(INT)m_LSpec_alldim;i++)
    free(delta[i]);
	free(delta);
	free(lfactor);
	free(factor);
  for (i=0;i<(INT)m_LSpec_DIMENSION;i++)
    free(dynamics[i]);
  free(dynamics);
	free(m_LSpec_found);
	free(m_LSpec_list);

 	for (i=0;i<BOXSPEC;i++)
    free(m_LSpec_box[i]);
  free(m_LSpec_box);
	free(m_LSpec_averr);
	free(var);
	free(av);
	free(interval);
	free(min);

	free(rnd_array);
	free(nexti);
	free(rnd1279);
	free(next1279);

	return LYAEXP;
}



void CLyapunovKantz::DrawExponents(CDC* pDC,HWND hwnd, double** LYAEXP, int count,int start,int dim,int tau)
{
	double TopYPlus=0;
	double TopYMinus=0;

	CRect rect;
	GetClientRect(hwnd,rect);

	for (int j=20;j<count;j++)
	{
		for (int i=0;i<(INT)dim;i++)
		{
			if ((LYAEXP[j][i])>TopYPlus) TopYPlus=LYAEXP[j][i];
			if ((LYAEXP[j][i])<TopYMinus) TopYMinus=LYAEXP[j][i];
		}
	}

	DOUBLE FaktorX,FaktorY;
	FaktorX=(DOUBLE)((rect.right-LEFTBORDER)/(FLOAT)start);
	FaktorY=(DOUBLE)((rect.bottom-BOTTOMBORDER-TOPBORDER)/(TopYPlus-TopYMinus));

	CPen RotStift,SchwarzStift,GruenStift,ExponentStift,BlauStift;
	ExponentStift.CreatePen(PS_DOT,1,RGB(0,0,0));
	RotStift.CreatePen(PS_SOLID,1,RGB(255,0,0));
	SchwarzStift.CreatePen(PS_SOLID,1,RGB(0,0,0));
	GruenStift.CreatePen(PS_SOLID,1,RGB(0,255,0));
	BlauStift.CreatePen(PS_SOLID,1,RGB(0,0,255));

	CAutoFont font("Century Gothic");
	font.SetDC(pDC->GetSafeHdc());
	font.SetHeight(16);
	font.SetFontColor(RGB(0,0,0));
	pDC->SelectObject(&font);

	if (m_oldFaktorY!=FaktorY) 
		pDC->FillSolidRect(0,0,rect.right,rect.bottom,RGB(255,255,255));
	else
		pDC->FillSolidRect(0,0,LEFTBORDER-1,rect.bottom,RGB(255,255,255));

	//Koordinatensystem zeichnen
	pDC->SelectObject(&SchwarzStift);
	//berschrift
	CString Text;
	Text.LoadString(IDS_LYAPUNOVSPEKTRUM);
 	pDC->DrawText(Text,rect,DT_TOP|DT_RIGHT);// TextOut(rect.right/2,16,Text);
	//X-Achse
	pDC->SelectObject(&SchwarzStift);
	Text.LoadString(IDS_ITERATIONEN);
	pDC->TextOut(rect.right-75,(LONG)(fabs(TopYPlus)*FaktorY+20+TOPBORDER),Text);
	pDC->MoveTo(5,(LONG)(fabs(TopYPlus)*FaktorY+TOPBORDER));
	pDC->LineTo(rect.right,(LONG)(fabs(TopYPlus)*FaktorY+TOPBORDER));
	for(j=0;j<(INT)start;j+=(MAX(start/5-(start/5)%100,20)))
	{
		if (j>0) 
		{
			pDC->MoveTo((LONG)(LEFTBORDER+j*FaktorX),(LONG)(fabs(TopYPlus)*FaktorY+TOPBORDER));
			pDC->LineTo((LONG)(LEFTBORDER+j*FaktorX),(LONG)(fabs(TopYPlus)*FaktorY+4+TOPBORDER));
			char buffer[20];
			CString strEvolve;
			strEvolve=itoa( j, buffer, 10 );
			pDC->TextOut((LONG)(LEFTBORDER-3+j*FaktorX),(LONG)(fabs(TopYPlus)*FaktorY+4+TOPBORDER),strEvolve);
		}
	}
	//Y-Achse
	pDC->MoveTo(LEFTBORDER,3);
	pDC->LineTo(LEFTBORDER,rect.bottom);

	//Exponenten
	for (int i=0;i<(INT)dim;i++)
	{
		pDC->MoveTo(LEFTBORDER,(LONG)((fabs(TopYPlus)-LYAEXP[0][i])*FaktorY+TOPBORDER));
		if (i==0) pDC->SelectObject(&RotStift);
		if (i==1) pDC->SelectObject(&BlauStift);
		if (i>1)  pDC->SelectObject(&GruenStift);

		DOUBLE Mittelwert=0;
		POINT Punkt;
		for (j=0;j<count;j++)
		{
			Punkt.x=(LONG)(LEFTBORDER+j*FaktorX);
			Punkt.y=(LONG)((fabs(TopYPlus)-LYAEXP[j][i])*FaktorY+TOPBORDER);
			if (Punkt.y<rect.bottom&&Punkt.y>=TOPBORDER) 
				pDC->LineTo(Punkt);
			else
			{
				if (Punkt.y>rect.bottom) 
					pDC->MoveTo(Punkt.x,rect.bottom);
				else 
					pDC->MoveTo(Punkt.x,TOPBORDER);
			}
			if (j==(INT)(start/2)) 
				Mittelwert=LYAEXP[j][i];
			if (j>start/2) 
				Mittelwert+=LYAEXP[j][i];
		}
		Mittelwert=Mittelwert/(count-(INT)(start/2));

		int dec;
		int sign;
		CString strWert=ecvt(Mittelwert,4, &dec,&sign );
		if (dec<=0) 
			for(int z=0;z<=abs(dec);z++) strWert.Insert(0,'0');
		if (dec<=0) strWert.Insert(1,'.');
		else strWert.Insert(dec,'.');
		if (sign!=0) strWert.Insert(0,'-');
		pDC->TextOut(2,(LONG)((fabs(TopYPlus)-Mittelwert)*FaktorY-8+TOPBORDER),strWert.Left(6));

		if (count==(INT)(start-dim*tau)) 
		{
			pDC->SelectObject(&ExponentStift);
			pDC->MoveTo(LEFTBORDER,(LONG)((fabs(TopYPlus)-Mittelwert)*FaktorY+TOPBORDER));
			pDC->LineTo(rect.right,(LONG)((fabs(TopYPlus)-Mittelwert)*FaktorY+TOPBORDER));
		}
	}

	m_oldFaktorY=FaktorY;
}



//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////
// Korrelationsdimension Tisean
//////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////

void CLyapunovKantz::scramble()
{
  long i,j,k,m;
  unsigned long rnd,rndf,hlength,allscr=0;
  long *scfound,*scnhelp,scnfound;
  long scbox[SCBOX],lswap,element,scbox1=SCBOX-1;
  double *rz,*schelp,sceps=(double)SCBOX-0.001,swap;
  
  hlength=m_Korr_length-(m_Korr_EMBED-1)*m_Korr_DELAY;

  if (sizeof(long) == 8) 
	{
    rndf=13*13*13*13;
    rndf=rndf*rndf*rndf*13;
    rnd=0x849178L;
  }
  else 
	{
    rndf=69069;
    rnd=0x234571L;
  }
  for (i=0;i<1000;i++)
    rnd=rnd*rndf+1;

  check_alloc(rz=(double*)malloc(sizeof(double)*hlength));
  check_alloc(scfound=(long*)malloc(sizeof(long)*hlength));
  check_alloc(scnhelp=(long*)malloc(sizeof(long)*hlength));
  check_alloc(schelp=(double*)malloc(sizeof(double)*hlength));

  for (i=0;i<(int)hlength;i++)
    rz[i]=(double)(rnd=rnd*rndf+1)/ULONG_MAX;
  
  for (i=0;i<SCBOX;i++)
    scbox[i]= -1;
  for (i=0;i<(int)hlength;i++) 
	{
    m=(int)(rz[i]*sceps)&scbox1;
    scfound[i]=scbox[m];
    scbox[m]=i;
  }
  for (i=0;i<SCBOX;i++) 
	{
    scnfound=0;
    element=scbox[i];
    while(element != -1) 
		{
      scnhelp[scnfound]=element;
      schelp[scnfound++]=rz[element];
      element=scfound[element];
    }
    
    for (j=0;j<scnfound-1;j++)
      for (k=j+1;k<scnfound;k++)
				if (schelp[k] < schelp[j]) 
				{
				  swap=schelp[k];
				  schelp[k]=schelp[j];
				  schelp[j]=swap;
				  lswap=scnhelp[k];
				  scnhelp[k]=scnhelp[j];
				  scnhelp[j]=lswap;
				}
    for (j=0;j<scnfound;j++)
      m_Korr_scr[allscr+j]=scnhelp[j];
    allscr += scnfound;
  }

  free(rz);
  free(scfound);
  free(schelp);
  free(scnhelp);
}

void CLyapunovKantz::make_c2_dim(int n)
{
  char smalle;
  long i,j,k,x,y,i1,i2,j1,element,n1,maxi,count,hi;
  double *hs,max,dx;
  
  check_alloc(hs=(double*)malloc(sizeof(double)*m_Korr_EMBED*m_Korr_DIM));
  n1=m_Korr_scr[n];

  count=0;
  for (i1=0;i1<(LONG)m_Korr_EMBED;i1++) 
	{
    i2=i1*m_Korr_DELAY;
    for (j=0;j<(LONG)m_Korr_DIM;j++)
      hs[count++]=m_Korr_Series[j][n1+i2];
  }

  x=(int)(hs[0]*m_Korr_epsinv)&m_Korr_imax;
  y=(int)(hs[1]*m_Korr_epsinv)&m_Korr_imax;
  
  for (i1=x-1;i1<=x+1;i1++) 
	{
    i2=i1&m_Korr_imax;
    for (j1=y-1;j1<=y+1;j1++) 
		{
      element=m_Korr_box[i2][j1&m_Korr_imax];
      while (element != -1) 
			{
				if (labs((long)(element-n1)) > (LONG)m_Korr_MINDIST) 
				{
				  count=0;
				  max=0.0;
				  maxi=m_Korr_howoften1;
				  smalle=0;
				  for (i=0;i<(int)m_Korr_EMBED;i++) 
					{
						hi=i*m_Korr_DELAY;
				    for (j=0;j<(LONG)m_Korr_DIM;j++) 
						{
							dx=fabs(hs[count]-m_Korr_Series[j][element+hi]);
							if (dx <= m_Korr_EPSMAX) 
							{
								if (dx > max) 
								{
									max=dx;
									if (max < m_Korr_EPSMIN) 
									{
								    maxi=m_Korr_howoften1;
									}
								  else 
									{
								    maxi=(LONG)((m_Korr_lneps-log(max))/m_Korr_lnfac);
									}
								}
								if (count > 0)
								  for (k=m_Korr_imin;k<=maxi;k++)
								    m_Korr_found[count][k] += 1.0;
							}
				      else 
							{
								smalle=1;
								break;
				      }
				      count++;
				    }
				    if (smalle)
				      break;
				  }
				}
				element=m_Korr_list[element];
      }
    }
  }

  free(hs);
}

void CLyapunovKantz::make_c2_1(int n)
{
  int i,x,i1,maxi;
  long element,n1;
  double hs,max;
  
  n1=m_Korr_scr[n];
  hs=m_Korr_Series[0][n1];
  
  x=(int)(hs*m_Korr_epsinv)&m_Korr_imax;
  
  for (i1=x-1;i1<=x+1;i1++) 
	{
    element=m_Korr_boxc1[i1&m_Korr_imax];
    while (element != -1) 
		{
      if (labs(element-n1) > (LONG)m_Korr_MINDIST) 
			{
				max=fabs(hs-m_Korr_Series[0][element]);
				if (max <= m_Korr_EPSMAX) 
				{
				  if (max < m_Korr_EPSMIN)
					 maxi=m_Korr_howoften1;
				  else
				    maxi=(LONG)((m_Korr_lneps-log(max))/m_Korr_lnfac);
				  for (i=m_Korr_imin;i<=maxi;i++)
				    m_Korr_found[0][i] += 1.0;
				}
      }
      element=m_Korr_listc1[element];
    }
  }
}


double*** CLyapunovKantz::Calc_Korr_Dim(CDC* pDC, HWND hwnd, double** Series, int Laenge)
{
	//////////////////////////////////////////////////////////////////////////
	//Nicht parametrisierte Variablen
	//////////////////////////////////////////////////////////////////////////
	char rescale_set=1;					//nicht normieren auf [0,1] sondern originaldaten verwenden
	unsigned long exclude=0;		//alle Inputdaten analysieren
	unsigned int verbosity=0x1;	//Fehlerausgabelevel wird nicht verwendet

	//lokale Variablen

	char dimset=0;
	char eps_min_set=1;					//Null, wenn MinRadius nicht gegeben
	char eps_max_set=1;					//Null, wenn MaxRadius nicht gegeben
//	m_Korr_EPSMAX=1.0;
//	m_Korr_EPSMIN=0.001;
	double epsfactor;
	double min;
	double interval;
	unsigned long nmax;

	double *norm;
  char smaller,stdi=0;
//  FILE *fout;
//	fout=fopen("C:\\Dumps\\KorrDim.txt","w");
	
  int maxembed;
  long i1,j1,x,y,sn,n,i,j,n1,n2;
  long *oscr;
  long lnorm;
  double eps,*epsm,EPSMAX1,maxinterval;
  time_t mytime,lasttime;
  
	double*** KorrDimWerte;
  check_alloc(KorrDimWerte=(double***)malloc(m_Korr_DIM*m_Korr_EMBED*sizeof(double**)*2));
  for (i=0;i<(LONG)(m_Korr_EMBED*m_Korr_DIM);i++)
	  check_alloc(KorrDimWerte[i]=(double**)malloc(m_Korr_HOWOFTEN*sizeof(double*)));
  for (i=0;i<(LONG)(m_Korr_EMBED*m_Korr_DIM);i++)
		for (j=0;j<(INT)m_Korr_HOWOFTEN;j++)
			check_alloc(KorrDimWerte[i][j]=(double*)malloc(2*sizeof(double)));



  m_Korr_Series=Series;
	m_Korr_length=Laenge;

  if (rescale_set) 
	{
    for (i=0;i<(LONG)m_Korr_DIM;i++)
      rescale_data(m_Korr_Series[i],m_Korr_length,&min,&interval);
    maxinterval=1.0;
  }
  else 
	{
    maxinterval=0.0;
    for (i=0;i<(LONG)m_Korr_DIM;i++) 
		{
      min=interval=m_Korr_Series[i][0];
      for (j=1;j<(LONG)m_Korr_length;j++) 
			{
				if (min > m_Korr_Series[i][j])
				  min=m_Korr_Series[i][j];
				if (interval < m_Korr_Series[i][j])
				  interval=m_Korr_Series[i][j];
      }
      interval -= min;
      if (interval > maxinterval)
				maxinterval=interval;
    }
  }
  if (!eps_max_set)
    m_Korr_EPSMAX *= maxinterval;
  if (!eps_min_set)
    m_Korr_EPSMIN *= maxinterval;
  m_Korr_EPSMAX=(fabs(m_Korr_EPSMAX)<maxinterval) ? fabs(m_Korr_EPSMAX) : maxinterval;
  m_Korr_EPSMIN=(fabs(m_Korr_EPSMIN)<m_Korr_EPSMAX) ? fabs(m_Korr_EPSMIN) :m_Korr_EPSMAX/2.;
  EPSMAX1=m_Korr_EPSMAX;

  m_Korr_howoften1=m_Korr_HOWOFTEN-1;
  maxembed=m_Korr_DIM*m_Korr_EMBED-1;

/*
  check_alloc(outd1=(char*)calloc(strlen(FOUT)+4,(size_t)1));
  strcpy(outd1,FOUT);
  strcat(outd1,".d2");
  check_alloc(outc1=(char*)calloc(strlen(FOUT)+4,(size_t)1));
  check_alloc(outh1=(char*)calloc(strlen(FOUT)+4,(size_t)1));
  check_alloc(outstat=(char*)calloc(strlen(FOUT)+6,(size_t)1));
  strcpy(outc1,FOUT);
  strcpy(outh1,FOUT);
  strcpy(outstat,FOUT);
  strcat(outc1,".c2");
  strcat(outh1,".h2");
  strcat(outstat,".stat");
*/


  check_alloc(m_Korr_list=(long*)malloc(m_Korr_length*sizeof(long)));
  check_alloc(m_Korr_listc1=(long*)malloc(m_Korr_length*sizeof(long)));
  if ((long)(m_Korr_length-(m_Korr_EMBED-1)*m_Korr_DELAY) <= 0) 
	{
    AfxMessageBox(E_LYAPUNOVKANTZ_CALCKORRDIM);
		g_bIsBreak=TRUE;
  }
  check_alloc(m_Korr_scr=(long*)malloc(sizeof(long)*(m_Korr_length-(m_Korr_EMBED-1)*m_Korr_DELAY)));
  check_alloc(oscr=(long*)malloc(sizeof(long)*(m_Korr_length-(m_Korr_EMBED-1)*m_Korr_DELAY)));
  check_alloc(m_Korr_found=(double**)malloc(m_Korr_DIM*m_Korr_EMBED*sizeof(double*)));
  for (i=0;i<(LONG)(m_Korr_EMBED*m_Korr_DIM);i++)
    check_alloc(m_Korr_found[i]=(double*)malloc(m_Korr_HOWOFTEN*sizeof(double)));
  check_alloc(norm=(double*)malloc(m_Korr_HOWOFTEN*sizeof(double)));
  check_alloc(epsm=(double*)malloc(m_Korr_HOWOFTEN*sizeof(double)));
  
  m_Korr_epsinv=1.0/m_Korr_EPSMAX;
  epsfactor=pow(m_Korr_EPSMAX/m_Korr_EPSMIN,1.0/(double)m_Korr_howoften1);
  m_Korr_lneps=log(m_Korr_EPSMAX);
  m_Korr_lnfac=log(epsfactor);

  epsm[0]=m_Korr_EPSMAX;
  norm[0]=0.0;
  for (i=1;i<(LONG)m_Korr_HOWOFTEN;i++) 
	{
    norm[i]=0.0;
    epsm[i]=epsm[i-1]/epsfactor;
  }
  m_Korr_imin=0;

  scramble();
  for (i=0;i<(LONG)(m_Korr_length-(m_Korr_EMBED-1)*m_Korr_DELAY);i++)
    oscr[m_Korr_scr[i]]=i;

  for (i=0;i<(LONG)(m_Korr_DIM*m_Korr_EMBED);i++)
    for (j=0;j<(LONG)m_Korr_HOWOFTEN;j++)
      m_Korr_found[i][j]=0.0;
  
  nmax=m_Korr_length-m_Korr_DELAY*(m_Korr_EMBED-1);

  for (i1=0;i1<NMAX;i1++) 
	{
    m_Korr_boxc1[i1]= -1;
    for (j1=0;j1<NMAX;j1++)
      m_Korr_box[i1][j1]= -1;
  }
  time(&lasttime);
  lnorm=0;
  
  for (n=1;n<(LONG)nmax;n++)  //nmax
	{
		//Abbruch mit ESC abfangen
		MSG msg;
		PeekMessage( &msg, NULL, 0, 0, PM_REMOVE );
    if (msg.message == WM_KEYDOWN && msg.wParam == VK_ESCAPE) 
		{
			g_bIsBreak=TRUE;
	    TranslateMessage(&msg); 
	    DispatchMessage(&msg); 
		}
		theApp.DoWaitCursor(0);
		if (g_bIsBreak) break;
    smaller=0;
    sn=m_Korr_scr[n-1];
    if (m_Korr_DIM > 1) 
		{
      x=(long)(m_Korr_Series[0][sn]*m_Korr_epsinv)&m_Korr_imax;
      y=(long)(m_Korr_Series[1][sn]*m_Korr_epsinv)&m_Korr_imax;
    }
    else 
		{
      x=(long)(m_Korr_Series[0][sn]*m_Korr_epsinv)&m_Korr_imax;
      y=(long)(m_Korr_Series[0][sn+m_Korr_DELAY]*m_Korr_epsinv)&m_Korr_imax;
    }
    m_Korr_list[sn]=m_Korr_box[x][y];
    m_Korr_box[x][y]=sn;
    m_Korr_listc1[sn]=m_Korr_boxc1[x];
    m_Korr_boxc1[x]=sn;

    i=m_Korr_imin;
    while (m_Korr_found[maxembed][i] >= m_Korr_MAXFOUND) 
		{
      smaller=1;
      if (++i > m_Korr_howoften1)
				break;
    }
    if (smaller) 
		{
      m_Korr_imin=i;
      if (m_Korr_imin <= m_Korr_howoften1) 
			{
				m_Korr_EPSMAX=epsm[m_Korr_imin];
				m_Korr_epsinv=1.0/m_Korr_EPSMAX;
				for (i1=0;i1<NMAX;i1++) 
				{
				  m_Korr_boxc1[i1]= -1;
				  for (j1=0;j1<NMAX;j1++)
						m_Korr_box[i1][j1]= -1;
				}
				for (i1=0;i1<n;i1++) 
				{
				  sn=m_Korr_scr[i1];
				  if (m_Korr_DIM > 1) 
					{
				    x=(long)(m_Korr_Series[0][sn]*m_Korr_epsinv)&m_Korr_imax;
						y=(long)(m_Korr_Series[1][sn]*m_Korr_epsinv)&m_Korr_imax;
				  }
				  else 
					{
				    x=(long)(m_Korr_Series[0][sn]*m_Korr_epsinv)&m_Korr_imax;
						y=(long)(m_Korr_Series[0][sn+m_Korr_DELAY]*m_Korr_epsinv)&m_Korr_imax;
				  }
				  m_Korr_list[sn]=m_Korr_box[x][y];
				  m_Korr_box[x][y]=sn;
				  m_Korr_listc1[sn]=m_Korr_boxc1[x];
				  m_Korr_boxc1[x]=sn;
				}
      }
    }

    if (m_Korr_imin <= m_Korr_howoften1) 
		{
      lnorm=n;
      if (m_Korr_MINDIST > 0) 
			{
				sn=m_Korr_scr[n];
				n1=(sn-(long)m_Korr_MINDIST>=0)?sn-(long)m_Korr_MINDIST:0;
				n2=(sn+m_Korr_MINDIST<m_Korr_length-(m_Korr_EMBED-1)*m_Korr_DELAY)?sn+m_Korr_MINDIST:
				  m_Korr_length-(m_Korr_EMBED-1)*m_Korr_DELAY-1;
				for (i1=n1;i1<=n2;i1++)
				  if ((oscr[i1] < n))
						lnorm--;
      }
      
      if (m_Korr_EMBED*m_Korr_DIM > 1)
				make_c2_dim(n);
      make_c2_1(n);
      for (i=m_Korr_imin;i<(LONG)m_Korr_HOWOFTEN;i++)
				norm[i] += (double)(lnorm);
    }
    
    if (((time(&mytime)-lasttime) > WHEN) || (n == (LONG)(nmax-1)) ||	(m_Korr_imin > m_Korr_howoften1)) 
		{
      time(&lasttime);

//      fprintf(fout,"#center= %ld\n",n);
      for (i=0;i<(LONG)(m_Korr_DIM*m_Korr_EMBED);i++) 
			{
//				fprintf(fout,"#dim= %ld\n",i+1);
				eps=EPSMAX1;
				for (j=1;j<(LONG)m_Korr_HOWOFTEN;j++) 
				{
				  eps /= epsfactor;
					double Wert=0;
				  if ((m_Korr_found[i][j] > 0.0) && (m_Korr_found[i][j-1] > 0.0))
					{
						Wert=log(m_Korr_found[i][j-1]/m_Korr_found[i][j]/norm[j-1]*norm[j])/m_Korr_lnfac;
//						fprintf(fout,"%e %e\n",eps,Wert);
						KorrDimWerte[i][j-1][0]=eps;
						KorrDimWerte[i][j-1][1]=Wert;
					}
					else
					{
						Wert=0.0f;
//						fprintf(fout,"%e %e\n",eps,Wert);
						KorrDimWerte[i][j-1][0]=eps;
						KorrDimWerte[i][j-1][1]=0;
					}
				}
//				fprintf(fout,"\n\n");
      }

	//////////////////////////////////////////////////////////////////////////
	//TISEAN AV-D2  This program takes the output files of the KorrDimWerte //
	//as its input and smoothes the content by arithmetically averaging over// 
	//a given interval.                                                     //
	//////////////////////////////////////////////////////////////////////////
	int aver=3;//*(MAX((INT)(m_Korr_HOWOFTEN/100.0),1)); //n average over (2n+1) values;
	double avy,aveps,norm;
	norm=2.0*aver+1.0;

	for (int i=0;i<(m_Korr_DIM*m_Korr_EMBED);i++)
	{
		for (int k=aver;k<m_Korr_HOWOFTEN-aver;k++) 
		{
		  avy=aveps=0.0;
			for (int j= -aver;j<=aver;j++) 
			{
				avy += KorrDimWerte[i][k+j][1];//y[k+j];
				aveps += KorrDimWerte[i][k+j][0];//eps[k+j];
			}
			if (fabs(aveps/norm)<1E50) KorrDimWerte[i][k][0]=aveps/norm;
			if (fabs(avy/norm)<1E50)		KorrDimWerte[i][k][1]=avy/norm;
		}
	}
	
	//////////////////////////////////////////////////////////////////////////

			DrawKorrDim(pDC,hwnd,KorrDimWerte,m_Korr_DIM*m_Korr_EMBED,m_Korr_HOWOFTEN,0,0);

      if (m_Korr_imin > m_Korr_howoften1)
				break;
				//				exit(0);
    }
  }
  
//	fclose(fout);

//	DrawKorrDim(pDC,hwnd,KorrDimWerte,m_Korr_DIM*m_Korr_EMBED,m_Korr_HOWOFTEN,0,0);

  
  free(m_Korr_list);
  free(m_Korr_listc1);
  free(m_Korr_scr);
  free(oscr);
  free(norm);
  free(epsm);
  for (i=0;i<(LONG)(m_Korr_EMBED*m_Korr_DIM);i++)
    free(m_Korr_found[i]);
  free(m_Korr_found);

/*
  for (i=0;i<m_Korr_DIM;i++)
    free(m_Korr_Series[i]);
  free(m_Korr_Series);
*/

  return KorrDimWerte;
}

//////////////////////////////////////////////////////////////////////////
// Einschub fr Linearen Fit nach Levenberg Marquart
//////////////////////////////////////////////////////////////////////////

#include "lmmin.h"
#include "lm_eval.h"
double ConstantFitFuncion( double t, double* p )
{
    return ( p[0] );
}
//////////////////////////////////////////////////////////////////////////

void CLyapunovKantz::DrawKorrDim(CDC* pDC,HWND hwnd, double*** KorrDimWerte,int AnzahlDimensionen,int AnzahlRadien,
																 LONG PosMarker1,LONG PosMarker2)
{
//	double TopYPlus=0;
//	double TopYMinus=0;

	CRect rect;
	GetClientRect(hwnd,rect);

/*	for (int i=0;i<AnzahlDimensionen;i++)
	{
		for (int j=0;j<AnzahlRadien-3;j++)
		{
			if (fabs(KorrDimWerte[i][j][1])<1E50) 
			{
				if ((KorrDimWerte[i][j][1])>TopYPlus) TopYPlus=KorrDimWerte[i][j][1];
				if ((KorrDimWerte[i][j][1])<TopYMinus) TopYMinus=KorrDimWerte[i][j][1];
			}
		}
	}
*/
	DOUBLE FaktorX,FaktorY;
	DOUBLE MaxX=1;//KorrDimWerte[0][0][0];
	DOUBLE MinX=KorrDimWerte[0][AnzahlRadien-2][0];
	FaktorX=(DOUBLE)((rect.right-LEFTBORDER)/fabs(log(MinX)));
	FaktorY=(DOUBLE)((rect.bottom-BOTTOMBORDER)/2/7);  //(TopYPlus-TopYMinus)); // /10

	CPen RotStift,SchwarzStift,GruenStift,ExponentStift,BlauStift;
	ExponentStift.CreatePen(PS_DOT,1,RGB(0,0,0));
	RotStift.CreatePen(PS_SOLID,1,RGB(255,0,0));
	SchwarzStift.CreatePen(PS_SOLID,1,RGB(0,0,0));
	GruenStift.CreatePen(PS_SOLID,1,RGB(0,255,0));
	BlauStift.CreatePen(PS_SOLID,1,RGB(0,0,255));

	CAutoFont font("Century Gothic");
	font.SetDC(pDC->GetSafeHdc());
	font.SetHeight(16);
	font.SetFontColor(RGB(0,0,0));
	pDC->SelectObject(&font);

	if (m_oldFaktorYKorr!=FaktorY) 
		pDC->FillSolidRect(0,0,rect.right,rect.bottom,RGB(255,255,255));
	else
		pDC->FillSolidRect(0,0,LEFTBORDER-1,rect.bottom,RGB(255,255,255));

	//Koordinatensystem zeichnen
	pDC->SelectObject(&SchwarzStift);
	//berschrift
	CString Text;
	Text.LoadString(IDS_KORRDIM);
 	pDC->DrawText(Text,rect,DT_TOP|DT_RIGHT);
	//X-Achse
	pDC->SelectObject(&SchwarzStift);
	Text.LoadString(IDS_RADIEN);
	pDC->TextOut(rect.right-150,(LONG)(rect.bottom-rect.bottom/2-18),Text);
	pDC->MoveTo(5,rect.bottom-rect.bottom/2-BOTTOMBORDER);
	pDC->LineTo(rect.right,rect.bottom-rect.bottom/2-BOTTOMBORDER);
	for(int j=1;j<=40;j++)
	{
		POINT Strich;
		double Wert;
		if (j<=10) Wert=(double)j/10000;
		else if (j<=20) Wert=((double)j-10)/1000;
					else if (j<=30) Wert=((double)j-20)/100;
								else if (j<=40) Wert=((double)j-30)/10;
	
		Strich.x=(LONG)(rect.right-fabs(log(Wert)*FaktorX));
		
		pDC->MoveTo(Strich.x,rect.bottom/2-BOTTOMBORDER);
		if (Strich.x>=LEFTBORDER) 
		{
			pDC->LineTo(Strich.x,rect.bottom/2-BOTTOMBORDER+4);
		}
		if (j%10==0)
		{
			int dec;
			int sign;
			CString strWert=ecvt(Wert,4, &dec,&sign );
			if (dec<=0) 
				for(int z=0;z<=abs(dec);z++) strWert.Insert(0,'0');
			if (dec<=0) strWert.Insert(1,'.');
			else strWert.Insert(dec,'.');
			if (sign!=0) strWert.Insert(0,'-');
			pDC->TextOut(Strich.x-8,rect.bottom/2-BOTTOMBORDER+5,strWert.Left(5));
		}
	}

	//Y-Achse
	font.SetEscapement(900);
	font.SetOrientation(900);
	pDC->SelectObject(&font);

	Text.LoadString(IDS_KORRDIM);
 	pDC->TextOut(4,rect.bottom/2-BOTTOMBORDER-10,Text);
	pDC->MoveTo(LEFTBORDER,0);
	pDC->LineTo(LEFTBORDER,rect.bottom-rect.bottom/2);

	font.SetEscapement(0);
	font.SetOrientation(0);
	pDC->SelectObject(&font);

	DOUBLE MaxDim=0;
	POINT LastPoint;
	LastPoint.x=LEFTBORDER;
	LastPoint.y=rect.bottom-BOTTOMBORDER;
	for(j=1;j<7;j++)
	{
		pDC->MoveTo(LEFTBORDER,(LONG)(rect.bottom-rect.bottom/2-BOTTOMBORDER-j*FaktorY));
		pDC->LineTo(LEFTBORDER-4,(LONG)(rect.bottom-rect.bottom/2-BOTTOMBORDER-j*FaktorY));
		char buffer[20];
		CString Dimension;
		Dimension=itoa( j, buffer, 10 );
		pDC->TextOut(LEFTBORDER-14,(LONG)(rect.bottom-rect.bottom/2-BOTTOMBORDER-j*FaktorY-8),Dimension);
	}


		//unteres System zeichnen
		//Trennlinie
		CPen TrennStift;
		TrennStift.CreatePen(PS_SOLID,3,RGB(0,0,0));
		pDC->SelectObject(&TrennStift);
		pDC->MoveTo(0,rect.bottom/2);
		pDC->LineTo(rect.right,rect.bottom/2);
		

		//X-Achse
		pDC->SelectObject(&SchwarzStift);
		pDC->MoveTo(5,rect.bottom-BOTTOMBORDER);
		pDC->LineTo(rect.right-5,rect.bottom-BOTTOMBORDER);
		for(j=1;j<=(INT)AnzahlDimensionen;j++)
		{
			pDC->MoveTo(LEFTBORDER+j*(rect.right-LEFTBORDER)/AnzahlDimensionen,rect.bottom-BOTTOMBORDER);
			pDC->LineTo(LEFTBORDER+j*(rect.right-LEFTBORDER)/AnzahlDimensionen,rect.bottom-BOTTOMBORDER+4);
			char buffer[20];
			CString Dimension;
			Dimension=itoa( j, buffer, 10 );
			pDC->TextOut(LEFTBORDER-3+j*(rect.right-LEFTBORDER)/AnzahlDimensionen,rect.bottom-BOTTOMBORDER+2,Dimension);
		}
		Text.LoadString(IDS_EINBETTUNGSDIMENSION);
		pDC->TextOut(rect.right-150,rect.bottom-16,Text);

		//Y-Achse
		for(j=1;j<6;j++)
		{
			pDC->MoveTo(LEFTBORDER,(LONG)(rect.bottom-BOTTOMBORDER-j*(rect.bottom-BOTTOMBORDER)/14));
			pDC->LineTo(LEFTBORDER-4,(LONG)(rect.bottom-BOTTOMBORDER-j*(rect.bottom-BOTTOMBORDER)/14));
			char buffer[20];
			CString Dimension;
			Dimension=itoa( j, buffer, 10 );
			pDC->TextOut(LEFTBORDER-14,(LONG)(rect.bottom-BOTTOMBORDER-8-j*(rect.bottom-BOTTOMBORDER)/14),Dimension);
		}
		font.SetEscapement(900);
		font.SetOrientation(900);
		pDC->SelectObject(&font);
		pDC->MoveTo(LEFTBORDER,rect.bottom/2+5);
		pDC->LineTo(LEFTBORDER,rect.bottom-5);
		Text.LoadString(IDS_KORRDIM);
		pDC->TextOut(4,rect.bottom-BOTTOMBORDER-10,Text);


	//Dimension zeichnen


	CString strMaxDim;
	CString strWert;

	for (int i=0;i<AnzahlDimensionen;i++)
	{
		if (g_bIsBreak) break;
		//oberes System
		font.SetEscapement(0);
		font.SetOrientation(0);
		pDC->SelectObject(&font);
		pDC->MoveTo(rect.right,(LONG)(rect.bottom-rect.bottom/2-BOTTOMBORDER-KorrDimWerte[i][0][1]*FaktorY));
//
//		CPen Stift;
//		COLORREF Farbe=PALETTEINDEX(i);
//		Stift.CreatePen(PS_SOLID,1,Farbe);
////		Stift.CreatePen(PS_SOLID,1,RGB(i/(FLOAT)AnzahlDimensionen*255,(AnzahlDimensionen-i)/(FLOAT)AnzahlDimensionen*255,50));//Farbe);
//		pDC->SelectObject(&Stift);

		if (i%3==0) pDC->SelectObject(&RotStift);
		if (i%3==1) pDC->SelectObject(&BlauStift);
		if (i%3==2) pDC->SelectObject(&GruenStift);

		DOUBLE Mittelwert=0;
		POINT Punkt;
		INT Count=0;
		CArray<DOUBLE,DOUBLE> RelevanteDimensionsWerte;
		CArray<DOUBLE,DOUBLE> RelevanteEpsilonWerte;
		for (j=0;j<AnzahlRadien-1;j++)
		{
			if (fabs(KorrDimWerte[i][j][1])<1E50) 
			{
				Punkt.x=(LONG)(rect.right-fabs(log(KorrDimWerte[i][j][0]))*FaktorX);
				Punkt.y=(LONG)(rect.bottom-rect.bottom/2-BOTTOMBORDER-KorrDimWerte[i][j][1]*FaktorY);
				if (Punkt.y>0&&Punkt.y<rect.bottom/2-BOTTOMBORDER) 
					pDC->LineTo(Punkt);
				else
				{
					if (Punkt.y>=rect.bottom/2-BOTTOMBORDER) 
						BOOL Nothing=TRUE;
					else 
						pDC->LineTo(Punkt.x,0);
				}
				if (Punkt.x>PosMarker1&&Punkt.x<PosMarker2) 
				{
					Mittelwert+=KorrDimWerte[i][j][1];
					RelevanteDimensionsWerte.Add(KorrDimWerte[i][j][1]);
					RelevanteEpsilonWerte.Add(KorrDimWerte[i][j][0]);
					Count++;
				}
			}
		}
		Mittelwert=Mittelwert/Count;

    // Fit initialisieren
    int m_dat = Count;  //Anzahl der Datenpunkte
    int n_p =  1;		 //Anzahl der Parameter

		//unteres System
		if (PosMarker1!=0&&PosMarker2!=0) 
		{
		  double* eps	 = RelevanteEpsilonWerte.GetData();
	    double* y		 = RelevanteDimensionsWerte.GetData();
			double  p[3] = { 1. }; // use any starting value, but not { 0 }

	    // auxiliary settings:
		  lm_control_type control;
		  lm_data_type data;
		  lm_initialize_control( &control );

		 data.user_func = ConstantFitFuncion;
		 data.user_t = eps;
	   data.user_y = y;
	
    // perform the fit:

//    AfxMessageBox( "modify or replace lm_print_default for less verbous fitting\n");
			lm_minimize( m_dat, n_p, p, lm_evaluate_default, lm_print_default, &data, &control );

    // print results:
//    printf( "status: %s after %d evaluations\n", lm_shortmsg[control.info], control.nfev );
			font.SetEscapement(0);
			font.SetOrientation(0);
			pDC->SelectObject(&font);
//			pDC->TextOut(rect.right-100,rect.bottom-BOTTOMBORDER-50,lm_shortmsg[control.info]);
//			pDC->TextOut(rect.right-20,rect.bottom-BOTTOMBORDER-50,IntToStr(control.nfev));

		
			int dec;
			int sign;
			strWert=ecvt(p[0],4, &dec,&sign );
			if (dec<=0) 
				for(int z=0;z<=abs(dec);z++) strWert.Insert(0,'0');
			if (dec<=0) strWert.Insert(1,'.');
			else strWert.Insert(dec,'.');
			if (sign!=0) strWert.Insert(0,'-');



			//Werte
			pDC->MoveTo(LEFTBORDER,rect.bottom-BOTTOMBORDER);
			pDC->SelectObject(&RotStift);

			pDC->MoveTo(LastPoint);
			POINT NeuPunkt;
			NeuPunkt.x=(LONG)(LEFTBORDER+(i+1)*(FLOAT)(rect.right-LEFTBORDER)/AnzahlDimensionen);
			NeuPunkt.y=(LONG)(rect.bottom-BOTTOMBORDER-p[0]*(rect.bottom-BOTTOMBORDER)/14);
			pDC->LineTo(NeuPunkt);
			if (p[0]>MaxDim)
			{
				strMaxDim=strWert.Left(6);
				MaxDim=p[0];
			}
			LastPoint=NeuPunkt;
			pDC->Ellipse(NeuPunkt.x-3,NeuPunkt.y-3,NeuPunkt.x+3,NeuPunkt.y+3);
			pDC->TextOut(NeuPunkt.x-20,NeuPunkt.y-20,strWert.Left(6));
			if (i==AnzahlDimensionen-1) 
			{
				pDC->SelectObject(&ExponentStift);
				LONG MaxDimY=(LONG)(rect.bottom-BOTTOMBORDER-MaxDim*(rect.bottom-BOTTOMBORDER)/14);
				pDC->MoveTo(LEFTBORDER,MaxDimY);
				pDC->LineTo(rect.right,MaxDimY);
				pDC->TextOut(LEFTBORDER+5,MaxDimY-17,strMaxDim);
			}
		}
	}
	m_oldFaktorY=FaktorY;
}


