
#include <math.h>

/*
 * Daubchies 16 nach Wickerhauser Anhang C
 *
 */
template <class T>
class Daubechies16 : public liftbase<T, double> {
  
protected:
  void predict( T& vec, int N, transDirection direction )
  {
    assert( false );
  } // predict
  
  virtual void update( T& vec, int N, transDirection direction )
  {
    assert( false );
  } // update
  
private:
  /** forward transform scaling coefficients */
  double h0, h1, h2, h3, h4, h5,h6,h7,h8,h9,h10,h11,h12,h13,h14,h15;
  /** forward transform wave coefficients */
  double g0, g1, g2, g3,g4,g5,g6,g7,g8,g9,g10,g11,g12,g13,g14,g15;
  
//  double Ih0, Ih1, Ih2, Ih3;
//  double Ig0, Ig1, Ig2, Ig3;

public:
  
  /**
    Forward Daubechies D16 transform step
    
    */
  void forwardStep( T& a, const int n )
  {
    if (n >= 16) {
      int i, j;
      const int half = n >> 1;
      
      double* tmp = new double[n];
      
      for (i = 0, j = 0; j < n-15; j += 2, i++) {
				tmp[i]      = a[j]*h0 + a[j+1]*h1 + a[j+2]*h2 + a[j+3]*h3 + a[j+4]*h4 + a[j+5]*h5 + a[j+6]*h6 + a[j+7]*h7+
								a[j+8]*h8 + a[j+9]*h9 + a[j+10]*h10 + a[j+11]*h11 + a[j+12]*h12 + a[j+13]*h13 + a[j+14]*h14 + 
								a[j+15]*h15;
				tmp[i+half] =a[j]*g0 + a[j+1]*g1 + a[j+2]*g2 + a[j+3]*g3 + a[j+4]*g4 + a[j+5]*g5 + a[j+6]*g6 + a[j+7]*g7+
								a[j+8]*g8 + a[j+9]*g9 + a[j+10]*g10 + a[j+11]*g11 + a[j+12]*g12 + a[j+13]*g13 + a[j+14]*g14 + 
								a[j+15]*g15;
      }
      
      tmp[i]      = a[n-14]*h0 + a[n-13]*h1 + a[n-12]*h2 + a[n-11]*h3 + a[n-10]*h4 + a[n-9]*h5 + a[n-8]*h6 +
										a[n-7]*h7+ a[n-6]*h8 + a[n-5]*h9 + a[n-4]*h10 + a[n-3]*h11 + a[n-2]*h12 + a[n-1]*h13 + 
										a[0]*h14 + 	a[1]*h15;
			
     tmp[i+half] = a[n-14]*g0 + a[n-13]*g1 + a[n-12]*g2 + a[n-11]*g3 + a[n-10]*g4 + a[n-9]*g5 + a[n-8]*g6 + 
									a[n-7]*g7+a[n-6]*g8 + a[n-5]*g9 + a[n-4]*g10 + a[n-3]*g11 + a[n-2]*g12 + a[n-1]*g13 + a[0]*g14 + 
								a[1]*g15;
      
      for (i = 0; i < n; i++) {
				a[i] = tmp[i];
      }
      delete [] tmp;
    }
  }
  
  /**
    Forward Daubechies D16 transform step, where the locations
    for the high and low pass filters are reversed.
    
    */
 void forwardStepRev( T& a, const int n )
  {
    if (n >= 16) {
      int i, j;
      const int half = n >> 1;
      
      double* tmp = new double[n];
      
      for (i = 0, j = 0; j < n-15; j += 2, i++) {
				tmp[i+half]      = a[j]*h0 + a[j+1]*h1 + a[j+2]*h2 + a[j+3]*h3 + a[j+4]*h4 + a[j+5]*h5 + a[j+6]*h6 + a[j+7]*h7+
								a[j+8]*h8 + a[j+9]*h9 + a[j+10]*h10 + a[j+11]*h11 + a[j+12]*h12 + a[j+13]*h13 + a[j+14]*h14 + 
								a[j+15]*h15;
				tmp[i] =a[j]*g0 + a[j+1]*g1 + a[j+2]*g2 + a[j+3]*g3 + a[j+4]*g4 + a[j+5]*g5 + a[j+6]*g6 + a[j+7]*g7+
								a[j+8]*g8 + a[j+9]*g9 + a[j+10]*g10 + a[j+11]*g11 + a[j+12]*g12 + a[j+13]*g13 + a[j+14]*g14 + 
								a[j+15]*g15;;
      }
      
      tmp[i+half]      = a[n-14]*h0 + a[n-13]*h1 + a[n-12]*h2 + a[n-11]*h3 + a[n-10]*h4 + a[n-9]*h5 + a[n-8]*h6 +
										a[n-7]*h7+ a[n-6]*h8 + a[n-5]*h9 + a[n-4]*h10 + a[n-3]*h11 + a[n-2]*h12 + a[n-1]*h13 + 
										a[0]*h14 + 	a[1]*h15;
			
     tmp[i] = a[n-14]*g0 + a[n-13]*g1 + a[n-12]*g2 + a[n-11]*g3 + a[n-10]*g4 + a[n-9]*g5 + a[n-8]*g6 + 
									a[n-7]*g7+a[n-6]*g8 + a[n-5]*g9 + a[n-4]*g10 + a[n-3]*g11 + a[n-2]*g12 + a[n-1]*g13 + a[0]*g14 + 
								a[1]*g15;
      
      for (i = 0; i < n; i++) {
				a[i] = tmp[i];
      }
      delete [] tmp;
    }
  }
  
  /**
    Inverse Daubechies D4 transform
    */
 /* void inverseStep( T& a, const int n )
  {
    if (n >= 4) {
      int i, j;
      const int half = n >> 1;
      const int halfPls1 = half + 1;
      
      double* tmp = new double[n];
      
      //      last smooth val  last coef.  first smooth  first coef
      tmp[0] = a[half-1]*Ih0 + a[n-1]*Ih1 + a[0]*Ih2 + a[half]*Ih3;
      tmp[1] = a[half-1]*Ig0 + a[n-1]*Ig1 + a[0]*Ig2 + a[half]*Ig3;
      for (i = 0, j = 2; i < half-1; i++) {
	//     smooth val     coef. val       smooth val    coef. val
	tmp[j++] = a[i]*Ih0 + a[i+half]*Ih1 + a[i+1]*Ih2 + a[i+halfPls1]*Ih3;
	tmp[j++] = a[i]*Ig0 + a[i+half]*Ig1 + a[i+1]*Ig2 + a[i+halfPls1]*Ig3;
      }
      for (i = 0; i < n; i++) {
	a[i] = tmp[i];
      }
      delete [] tmp;
    }
  } // inverseStep
*/  

  /**
    Initialize the filter constants used in the Daubechies 
    D16 transform.
   */
  Daubechies16() 
  {
    //
    // forward transform scaling (smoothing) coefficients
    //
    h0 = -0.000117476784;
    h1 = -0.000675449406;
    h2 = -0.000391740373;
    h3 = 0.004870352993;
		h4 = 0.008746094047;
		h5 = -0.013981027917;
		h6 = -0.044088253931;
		h7 = 0.017369301002;
		h8 = 0.12874742662;
		h9 = -0.000472484574;
		h10= -0.284015542962;
		h11= 0.015829105256;
		h12= 0.585354683654;
		h13= -0.675630736297;
		h14= 0.312871590914;
		h15= -0.054415842243;

    //
    // forward transform wavelet coefficients (a.k.a. high
    // pass filter coefficients)
    //
    g0 = -h15;
    g1 = h14;
    g2 = -h13;
    g3 = h12;
		g4 = -h11;
		g5 = h10;
		g6 = -h9;
		g7 = h8;
		g8 = -h7;
		g9 = h6;
		g10= -h5;
		g11= h4;
		g12= -h3;
		g13= h2;
		g14= -h1;
		g15= h0;
    
 /*   Ih0 = h2;
    Ih1 = g2;  // h1
    Ih2 = h0;
    Ih3 = g0;  // h3
    
    Ig0 = h3;
    Ig1 = g3;  // -h0
    Ig2 = h1;
    Ig3 = g1;  // -h2
 */
	}

  /**
    Forward Daubechies D16 transform
   */
  void forwardTrans( T& ts, int N )
  {
    int n;
    for (n = N; n >= 16; n >>= 1) {
      forwardStep( ts, n );
    }
  } // forwardTrans
  
  
  /**
    Inverse Daubechies D16 transform
   */
/*  void inverseTrans( T& coef, int N )
  {
    int n;
    for (n = 16; n <= N; n <<= 1) {
      inverseStep( coef, n );
    }
  } // inverseTrans
  */
}; // Daubechies16

