// wavelets.cpp

#include "stdafx.h"
#include "wavelets.h"

#define ALGORITHM_ZERO_PAD_GENERIC 0
#define ALGORITHM_ZERO_PAD_OPTIMIZED 1
#define ALGORITHM_MIRROR_GENERIC 2
#define ALGORITHM_ZERO_PAD_HISTORY_GENERIC 3
#define ALGORITHM_NUMERICAL_RECIPES 4
#define ALGORITHM_COUNT 5;

// forward declarations
void wt1(double a[], double worksp[], unsigned long n, int isign, wavefilt* wfilt,
		 void (*wtstep)(double [], double[], unsigned long, int, wavefilt*));
void pwt(double a[], double worksp[], unsigned long n, int isign, wavefilt* wfilt);

inline int getMaxLevelCount(int totalSignalLength) {
	return (int) (((float) (log(totalSignalLength)/log(2)))-1);
//	return (int) (Round((float) (log(totalSignalLength)/log(2)))-1);
}

WaveletCoeffs::WaveletCoeffs() {
	m_coeffBuffer=NULL;
	m_coeffBufferSize=0;
	reset(0,0);
}

WaveletCoeffs::~WaveletCoeffs() {
	if (m_coeffBuffer!=NULL) {
		delete[] m_coeffBuffer;
		m_coeffBuffer=NULL;
		m_coeffBufferSize=0;
	}
}

void WaveletCoeffs::internalReset(int totalSignalLength, int bufferSize) {
	m_totalSignalLength=totalSignalLength;
	if ((m_coeffBuffer==NULL || m_coeffBufferSize<bufferSize) && totalSignalLength>0) {
		if (m_coeffBuffer!=NULL) {
			delete[] m_coeffBuffer;
			m_coeffBuffer=NULL;
			m_coeffBufferSize=0;
		}
		m_coeffBuffer=new SAMPLE[bufferSize];
		m_coeffBufferSize=bufferSize;
	}
	m_approx=NULL;
	m_approxCount=0;
	m_coeffs.RemoveAll();
	m_counts.RemoveAll();
	setLinearMemory(NULL, 0);
}

void WaveletCoeffs::reset(int totalSignalLength, int filterLength) {
	int bufSize=getMaxLevelCount(totalSignalLength)*2*filterLength+totalSignalLength;
	internalReset(totalSignalLength, bufSize);
}

void WaveletCoeffs::addLevel(SAMPLE* coeffs, int count) {
	m_coeffs.Add(coeffs);
//	this->coeffs.Add(*coeffs);
	m_counts.Add( count);
//	counts.AddTail((void*) count);
}

void WaveletCoeffs::setApprox(SAMPLE* coeffs, int count) {
	m_approx=coeffs;
	m_approxCount=count;
}

int WaveletCoeffs::getLevelCount() {
	return m_coeffs.GetSize();
}

SAMPLE* WaveletCoeffs::getDetails(int level, int* count) {
	if (level<0 || level>=getLevelCount()) {
		AfxMessageBox("WaveletCoeffs::getDetails(%d, count): out of range.");
//		EH_error(2, "WaveletCoeffs::getDetails(%d, count): out of range.", level);
		return NULL;
	}
	*count=m_counts.GetAt(level);
	return m_coeffs.GetAt(level);
}

int WaveletCoeffs::getDetailCount(int level) { 
	if (level>=0 && level<m_counts.GetSize()) {
		return (int) m_counts.GetAt(level);
	} else
	if (level<0) {
		return m_totalSignalLength;
	} else
	return 0;
}

SAMPLE* WaveletCoeffs::getApprox(int* count) {
	*count=m_approxCount;
	return m_approx;
}

void WaveletCoeffs::copyFrom(WaveletCoeffs* otherCoeffs) {
	internalReset(otherCoeffs->getTotalSignalLength(), otherCoeffs->getCoeffBufferSize());
	// copy buffer and just map the pointers. Should be faster
	// than going through each level and copying each level
	// after another
	SAMPLE* otherBuffer=otherCoeffs->getCoeffBuffer();
	int count=otherCoeffs->getCoeffBufferSize();
	// copy the entire buffer
	memcpy(m_coeffBuffer, otherBuffer, sizeof(SAMPLE)*count);
	// now map all the coefficients
	DWORD diff=((DWORD) otherBuffer)-((DWORD) m_coeffBuffer);
	count=otherCoeffs->getLevelCount();
	int thisCount;
	DWORD otherElement;
	for (int i=0; i<count; i++) {
		otherElement=(DWORD) (otherCoeffs->getDetails(i, &thisCount));
		addLevel((SAMPLE*) (otherElement-diff), thisCount);
	}
	otherElement=(DWORD) (otherCoeffs->getApprox(&thisCount));
	setApprox((SAMPLE*) (otherElement-diff), thisCount);
	if (otherCoeffs->isLinearMemory()) {
		otherElement=(DWORD) (otherCoeffs->getLinearMemory());
		setLinearMemory((SAMPLE*) (otherElement-diff), otherCoeffs->getLinearMemoryCount());
	}
}

//////////////////////////// WaveletHistory ////////////////////////////////

WaveletHistory::WaveletHistory() {
	m_currLevelCount=0;
}

WaveletHistory::~WaveletHistory() {
	destroyMemory();
}

void WaveletHistory::reset() { // sets all history values to 0
	for (int i=0; i<m_hcoeffs.GetSize(); i++) {
		reset(i);
	}
}

void WaveletHistory::reset(int level) {
	SAMPLE* curr=getHistory(level);
	for (int i=MAXIMUM_FILTER_LENGTH; i>0; i--) {
		*curr=0;
	}
	m_hcounts.SetAt(level, (int*) 0);
}

void WaveletHistory::moveSamples(int level, SAMPLE* coeffs, int coeffCount, int count) {
	SAMPLE* curr=getHistory(level);
	int prevHistoryCount=(int) m_hcounts.GetAt(level);
	int newHistoryCount=count;
	if (count<=0 || coeffCount<=0) {
		if (prevHistoryCount>0) {
			reset(level);
		} else {
			m_hcounts.SetAt(level,(int*) 0);
		}
		return;
	}

	if (coeffCount<count) {
		if (prevHistoryCount>0 && coeffCount<count) {
			// need to move the last entries (when filter length is longer than level)
			int thisCount=count-coeffCount;
			int offset=coeffCount;
			for (int i=0; i<thisCount; i++) {
				curr[i+offset]=curr[i];
			}
			newHistoryCount=prevHistoryCount+coeffCount;
			if (newHistoryCount>count) {
				newHistoryCount=count;
			}
		} else {
			newHistoryCount=coeffCount;
		}
		count=coeffCount;
	}
	m_hcounts.SetAt(level, (int*) newHistoryCount);
	// finally copy old samples to history
	coeffs+=coeffCount;
	for (int i=0; i<count; i++) {
		coeffs--;
		*curr=*coeffs;
		curr++; 
	}
}

void WaveletHistory::update(WaveletCoeffs* coeffs, int filterLength) {
	// copy the last <<filterlength-2>> coeffs to the history
	int length=coeffs->getTotalSignalLength();
	int coeffCount;
	SAMPLE* coe;
	assertLevels(coeffs->getLevelCount());
	for (int level=1; level<coeffs->getLevelCount(); level++) {
		length>>=1;
		coe=coeffs->getDetails(level-1, &coeffCount);
		// take the middle part of the coeffs
		coe+=((coeffCount-length)>>1);
		moveSamples(level, coe, length, filterLength-2);
	}
}

void WaveletHistory::updateSignal(SAMPLE* signal, int signalLength, int filterLength) {
	// copy the last <<filterlength-2>> samples to the history level 0
	assertLevels(1);
	moveSamples(0, signal, signalLength, filterLength-2);
}

SAMPLE* WaveletHistory::getHistory(int level) {
#ifdef _DEBUG
	if (level<0 || level>=m_hcoeffs.GetSize()) {
		AfxMessageBox("Wrong call to WaweletHistory::getHistory(int): level=%d out of range.");
//		EH_debug(1,"Wrong call to WaweletHistory::getHistory(int): level=%d out of range.", level);
		level=0;
		assertLevels(1);
	}
#endif
	return (SAMPLE*) m_hcoeffs.GetAt(level);
}

SAMPLE* WaveletHistory::getHistoryNoCheck(int level) {
	if (level<0 || level>=m_hcoeffs.GetSize()) {
		return NULL;
	}
	return (SAMPLE*) m_hcoeffs.GetAt(level);
}

void WaveletHistory::destroyMemory() {
	for (int i=0; i<m_hcoeffs.GetSize(); i++) {
		SAMPLE* curr=getHistory(i);
		delete[] curr;
	}
	m_hcoeffs.RemoveAll();
	m_hcounts.RemoveAll();
	m_currLevelCount=0;
}

void WaveletHistory::assertLevels(int levels) {
	while (m_hcoeffs.GetSize()<levels) {
		SAMPLE* newLevel=new SAMPLE[MAXIMUM_FILTER_LENGTH];
		m_hcoeffs.Add( newLevel);
		m_hcounts.Add( (int*)0);
		reset(m_hcoeffs.GetSize()-1);
	}
}

/////////////////////////// WaveletTransform ////////////////////////////////

WaveletTransform::WaveletTransform() {
	m_currWavelet=1;
	m_algorithm=4;
	setWavelet(0);
	setAlgorithm(4);
	m_workingBuffer=NULL;
	m_workingBufferSize=0;
	initWaveletFilterCoeffs();
}

WaveletTransform::~WaveletTransform() {
	delWorkingBuffer();
}

void WaveletTransform::assertWorkingBuffer(int totalSignalLength) {
	int bufSize=getMaxLevelCount(totalSignalLength)*2*waveletFilterCoeffs[m_currWavelet].lowPassForwardLength+totalSignalLength;
	if (m_workingBufferSize<(bufSize*2)) {
		if (m_workingBuffer!=NULL) {
			delete[] m_workingBuffer;
			m_workingBuffer=NULL;
			m_workingBufferSize=0;
		}
		// allocate enough memory for wavelet coefficients
		m_workingBuffer=new SAMPLE[bufSize*2];
		m_workingBufferSize=bufSize*2;
	}
}

void WaveletTransform::delWorkingBuffer() {
	if (m_workingBuffer!=NULL) {
		delete[] m_workingBuffer;
		m_workingBuffer=NULL;
		m_workingBufferSize=0;
	}
}

// returns outCount for output and details
int forwardFilterZeroPadGeneric(SAMPLE* input, int inCount, SAMPLE* details, SAMPLE* approx,
											float* lowFilter, float* highFilter, int filterLength, SAMPLE* history) {
	int paddedSignalLength=2*(filterLength-2)+inCount;
	int outCount=(1+paddedSignalLength-(filterLength-2)) >> 1; // div 2

	// it starts negative: zero padding before signal
	int inputStart=-(filterLength-2);
	// zero padding after signal
	int inputEnd=filterLength-2+inCount;
	
	int inIndex=inputStart;
	for (int outIndex=0; outIndex<outCount; outIndex++) {
		approx[outIndex]=0;
		details[outIndex]=0;
		// convolution
		for (int i=0; i<filterLength; i++) {
			if (i+inIndex>=0 && i+inIndex<inCount) {
				// if the filter "hits" the real signal
				approx[outIndex]+=lowFilter[i]*input[i+inIndex];
				details[outIndex]+=highFilter[i]*input[i+inIndex];
			}
		}
		inIndex+=2;
	}
	return outCount;
}

int forwardFilterZeroPadOptimized(SAMPLE* input, int inCount, SAMPLE* details, SAMPLE* approx,
											float* lowFilter, float* highFilter, int filterLength, SAMPLE* history) {
	//if (inCount%2==1)
	//return 
	//	forwardFilterZeroPadGeneric(input, inCount, details, approx,
	//										lowFilter, highFilter, filterLength, history);

	// resulting SAMPLES per approx and details
	int outCount=0;
	SAMPLE* det=details;
	SAMPLE* apr=approx;
	
	// padding samples before and after input: filterLength-2
	int pads=filterLength-2;

	// "left" edge
	for (int i=0; i<pads; i++, i++) {
		SAMPLE* innerInput=input;
		*apr=0;
		*det=0;
		int max=i+2;
		if (max>inCount) {
			max=inCount;
		}
		max+=pads-i;
		for (int inner=pads-i; inner<max; inner++) {
			*apr+=*innerInput*lowFilter[inner];
			*det+=*innerInput*highFilter[inner];
			innerInput++;
		}
		apr++;
		det++;
		outCount++;
	}
	//int test=0;
	//int maxInput=0;

	// middle area (full filter) 
	// if inCount is uneven, the uneven sample is handled in next loop
	for (i=inCount-pads-1; i>0; i--, i--) {
		SAMPLE* innerInput=input;
		*apr=0;
		*det=0;
		for (int inner=0; inner<filterLength; inner++) {
			/*if (test+inner>=inCount) {
				EH_debug(-1,"1 test+inner=%d inCount=%d filterLen=%d i=%d", test+inner, inCount, filterLength, i);
			}*/
			*apr+=*innerInput*lowFilter[inner];
			*det+=*innerInput*highFilter[inner];
			innerInput++;
			//if (test+inner>maxInput) maxInput=test+inner;
		}
		apr++;
		det++;
		input++; input++;
		//test+=2;
		outCount++;
	}

	// right area (left coefficients of filter only)
	int max=pads;
	if (inCount<pads) {
		max=inCount;
	} else 
	if ((inCount & 1)==1) {
		// uneven number of input samples: handle the last sample !
		max++;
	}
	for (i=max; i>0; i--, i--) {
		SAMPLE* innerInput=input;
		*apr=0;
		*det=0;
		for (int inner=0; inner<i; inner++) {
			*apr+=*innerInput*lowFilter[inner];
			*det+=*innerInput*highFilter[inner];
			//if (test+inner>maxInput) maxInput=test+inner;
			//if (test+inner>=inCount) {
				//EH_debug(-1,"2 test+inner=%d inCount=%d max=%d filterLen=%d i=%d", test+inner, inCount, max, filterLength, i);
			//}
			innerInput++;
		}
		apr++;
		det++;
		input++; input++;
		//test+=2;
		outCount++;
	}
	//if (maxInput!=inCount-1) {
	//	EH_debug(-1,"3 maxInput=%d inCount=%d filterLen=%d", maxInput, inCount, filterLength);
	//}
	//int paddedSignalLength=2*(filterLength-2)+inCount;
	//int refOutCount=(1+paddedSignalLength-(filterLength-2)) >> 1; // div 2
	//if (refOutCount!=outCount) {
	//	EH_debug(-1,"4 outCount wrong ! refOutCount=%d outCount=%d filterLen=%d", refOutCount, outCount, filterLength);
	//}
	return outCount;
}

// returns outCount for output and details
int forwardFilterMirrorGeneric(SAMPLE* input, int inCount, SAMPLE* details, SAMPLE* approx,
											float* lowFilter, float* highFilter, int filterLength, SAMPLE* history) {
	int paddedSignalLength=2*(filterLength-2)+inCount;
	int outCount=(1+paddedSignalLength-(filterLength-2)) >> 1; // div 2

	// negative: zero padding before signal
	int inputStart=-(filterLength-2);
	// zero padding after signal
	int inputEnd=filterLength-2+inCount;
	
	int inIndex=inputStart;
	int mirroredInIndex;
	for (int outIndex=0; outIndex<outCount; outIndex++) {
		approx[outIndex]=0;
		details[outIndex]=0;
		// convolution
		for (int i=0; i<filterLength; i++) {
			mirroredInIndex=i+inIndex;
			if (mirroredInIndex>=inCount) {
				mirroredInIndex=inCount-(mirroredInIndex-inCount+1);
			} else
			if (mirroredInIndex<0) {
				mirroredInIndex=-mirroredInIndex;
			}
			if (mirroredInIndex>=0 && mirroredInIndex<inCount) {
				// if the filter "hits" the real signal
				approx[outIndex]+=lowFilter[i]*input[mirroredInIndex];
				details[outIndex]+=highFilter[i]*input[mirroredInIndex];
			}
		}
		inIndex+=2;
	}
	return outCount;
}

int forwardFilterHistoryGeneric(SAMPLE* input, int inCount, SAMPLE* details, SAMPLE* approx,
											float* lowFilter, float* highFilter, int filterLength, SAMPLE* history) {
	int paddedSignalLength=2*(filterLength-2)+inCount;
	int outCount=(1+paddedSignalLength-(filterLength-2)) >> 1; // div 2

	// negative: zero padding before signal
	int inputStart=-(filterLength-2);
	// zero padding after signal
	int inputEnd=filterLength-2+inCount;
	
	int inIndex=inputStart;
	int thisInIndex;
	for (int outIndex=0; outIndex<outCount; outIndex++) {
		approx[outIndex]=0;
		details[outIndex]=0;
		// convolution
		for (int i=0; i<filterLength; i++) {
			thisInIndex=i+inIndex;
			if (thisInIndex<0 && history!=NULL) {
				// use history
				approx[outIndex]+=lowFilter[i]*history[-thisInIndex-1];
				details[outIndex]+=highFilter[i]*history[-thisInIndex-1];
			} else {
				if (thisInIndex>=inCount) {
					thisInIndex=inCount-(thisInIndex-inCount+1);
				} else
				if (thisInIndex<0) {
					thisInIndex=-thisInIndex;
				}
				if (thisInIndex>=0 && thisInIndex<inCount) {
					// if the filter "hits" the real signal
					approx[outIndex]+=lowFilter[i]*input[thisInIndex];
					details[outIndex]+=highFilter[i]*input[thisInIndex];
				}
			}
		}
		inIndex+=2;
	}
	return outCount;
}

void inverseFilterGeneric(SAMPLE* output, SAMPLE* details, SAMPLE* approx, int coeffCount,
								int outCount,
								float* lowFilter, float* highFilter, int filterLength) {
	// only calculate non-paddings
	int inIndex=0;
	for (int outIndex=0; outIndex<outCount; outIndex+=2) {
		output[outIndex]=0;
		output[outIndex+1]=0;
		int i=inIndex;
		for (int filterIndex=0; filterIndex<filterLength; filterIndex+=2) {
			output[outIndex]+=lowFilter[filterIndex+1]*approx[i]+highFilter[filterIndex+1]*details[i];
			if (outIndex+1<outCount) {
				output[outIndex+1]+=lowFilter[filterIndex]*approx[i]+highFilter[filterIndex]*details[i];
			}
			i++;
		}
		inIndex++;
	}
}

void inverseFilterOptimized(SAMPLE* output, SAMPLE* details, SAMPLE* approx, int coeffCount,
								int outCount,
								float* lowFilter, float* highFilter, int filterLength) {
	// only calculate non-paddings
	SAMPLE* output0=output;
	SAMPLE* output1=output; output1++;
	int inIndex=0;
	for (int ii=outCount-1; ii>0; ii--, ii--) {
		SAMPLE* innerApprox=approx;
		SAMPLE* innerDetails=details;
		// first iteration of the convolution is executed directly here
		*output0=lowFilter[1]*(*innerApprox)+highFilter[1]*(*innerDetails);
		*output1=*lowFilter*(*innerApprox)+(*highFilter)*(*innerDetails);
		for (int filterIndex=2; filterIndex<filterLength; filterIndex++) {
			innerApprox++;
			innerDetails++;
			*output1+=lowFilter[filterIndex]*(*innerApprox)+highFilter[filterIndex]*(*innerDetails);
			filterIndex++;
			*output0+=lowFilter[filterIndex]*(*innerApprox)+highFilter[filterIndex]*(*innerDetails);
		}
		output0++; output0++;
		output1++; output1++;
		approx++; details++;
	}
	// if uneven number of output samples, add one
	if (outCount & 1==1) {
		SAMPLE* innerApprox=approx;
		SAMPLE* innerDetails=details;
		// increment low and high pass filter, eleminates need for "+1" in the index
		lowFilter++; highFilter++;
		// first iteration of the convolution is executed directly here
		*output0=*lowFilter*(*innerApprox)+(*highFilter)*(*innerDetails);
		for (int filterIndex=2; filterIndex<filterLength; filterIndex++, filterIndex++) {
			innerApprox++;
			innerDetails++;
			*output0+=lowFilter[filterIndex]*(*innerApprox)+highFilter[filterIndex]*(*innerDetails);
		}
	}
}


void WaveletTransform::numericalRecipesForwardTransform(SAMPLE* input, int inputLength, WaveletCoeffs* coeffs) {
	// copy input to coeffBuffer. wt1 does quasi-inplace transform
	SAMPLE* coeffBuffer=coeffs->getCoeffBuffer();
	for (int i=0; i<inputLength; i++) {
		coeffBuffer[i]=input[i];
	}
	// -1 because NR uses 1-based arrays
	wt1(coeffBuffer-1,m_workingBuffer-1, inputLength, 1, &m_wfilt, *pwt);
	
	// fill our own WaveletCoeffs class
	SAMPLE* temp=coeffBuffer;
	temp+=inputLength;
	int nn=inputLength; nn>>=1;
	for (; nn>=2; nn>>=1) {
		temp-=nn;
		coeffs->addLevel(temp, nn);
	}
	coeffs->setApprox(coeffBuffer, 2);
	// support optimization
	coeffs->setLinearMemory(coeffBuffer, inputLength);
}


void WaveletTransform::numericalRecipesInverseTransform(WaveletCoeffs* input, SAMPLE* output) {
	// probably input lies in our own coeffBuffer area.
	// but it is not sure. So create a copy
	SAMPLE* temp=output;
	int count;
	SAMPLE* c=input->getApprox(&count);
	
	// assumption: full level decomposition

	// copy approx level
	*temp=*c; temp++; c++;
	*temp=*c; temp++; c++;
	
	// copy detail level
	int unpaddedCount=2;
	for (int level=input->getLevelCount()-1; level>=0; level--, unpaddedCount<<=1) {
		c=input->getDetails(level, &count);
		if (count>unpaddedCount) {
			count=unpaddedCount;
		}
		for (int i=0; i<count; i++, c++, temp++) {
			*temp=*c;
		}
	}
	wt1(output-1, m_workingBuffer-1, input->getTotalSignalLength(), -1, &m_wfilt, *pwt);
}



// without wrap (and more coefficients output)
void WaveletTransform::generalForwardTransform(SAMPLE* input, int inputLength, WaveletCoeffs* coeffs, WaveletHistory* history) {
	WaveletFilterCoeffs* params=&(waveletFilterCoeffs[m_currWavelet]);
	SAMPLE* detailOut=coeffs->getCoeffBuffer();
	SAMPLE* currIn=input;
	SAMPLE* approxOut=m_workingBuffer;
	int currOutCount;
	int currInCount=inputLength;
	int level=0;
	for (int counter=inputLength; counter>=4; counter>>=1, level++) {
		currOutCount=(*m_forwardFunc)(currIn, currInCount, detailOut, approxOut, 
			params->lowPassFlipped, params->highPassFlipped, params->lowPassForwardLength, 
			(history!=NULL)?history->getHistoryNoCheck(level):NULL);
		if (currOutCount>0) {
			coeffs->addLevel(detailOut, currOutCount);
			detailOut+=currOutCount;
		} else { 
			break;
		}

/*		//Debuging Approx
		char a[1000];
		a[0]=0;
		for (int i=0; i<currOutCount; i++) 
		{
//		sprintf(a, "%s %.1f, ", a, approxOut[i]);
		}
		CString Ausgabe;
		sprintf(Ausgabe.GetBuffer(5000),"In ForwardTransform: level %d, %d approx coeffs= %s", coeffs->getLevelCount()-1, currOutCount, a);
		AfxMessageBox(Ausgabe);
//		EH_debug(-1, "In ForwardTransform: level %d, %d approx coeffs= %s", coeffs->getLevelCount()-1, currOutCount, a);
*/

		// approx level becomes new input
		currIn=approxOut;
		currInCount=currOutCount;
		// don't overwrite input samples with approximation output in working buffer
		if (approxOut==m_workingBuffer) {
			approxOut+=currOutCount;
		} else {
			approxOut=m_workingBuffer;
		}
//break; // testing: only one level
	}
	// the last lowpass remainder is copied to coeffBuffer
	memcpy(detailOut, currIn, currInCount*sizeof(SAMPLE));
	// and add it to the WaveletCoeffs as approx coeffs
	coeffs->setApprox(detailOut, currInCount);
	// now update history, if necessary
	if (history!=NULL && m_algorithm==ALGORITHM_ZERO_PAD_HISTORY_GENERIC) {
		history->update(coeffs, params->lowPassForwardLength);
		history->updateSignal(input, inputLength, params->lowPassForwardLength);
	}
}


void WaveletTransform::generalInverseTransform(WaveletCoeffs* input, SAMPLE* output) {
	int levelCount=input->getLevelCount();
	int approxCount;
	int detailCount;
	int outCount;
	SAMPLE* approx=input->getApprox(&approxCount);
	SAMPLE* tempOutput;
	WaveletFilterCoeffs* params=&(waveletFilterCoeffs[m_currWavelet]);
	for (int level=(levelCount-1); level>=0; level--) {
		// every second iteration will use the real output buffer.
		// especially the last iteration will write directly to output
		if ((level & 1)==0) {
			tempOutput=output;
		} else {
			// write the output of the inverse filter to working buffer
			tempOutput=m_workingBuffer;
		}

		SAMPLE* details=input->getDetails(level, &detailCount);
		outCount=input->getDetailCount(level-1);
		(*m_inverseFunc)(tempOutput, details, approx, detailCount,
			outCount,
			params->lowPassInverseFlipped, params->highPassInverseFlipped, params->lowPassForwardLength);
		
		// in next level, the current output will be the approximation
		approx=tempOutput;
		approxCount=outCount;
/*char a[5000];
a[0]=0;
for (int i=0; i<outCount; i++) {
	sprintf(a, "%s %.4f, ", a, tempOutput[i]);
}
EH_debug(-1, "In InverseTransform: level %d, %d approx coeffs= %s", level-1, outCount, a);
*/
	}
}

void WaveletTransform::forwardTransform(SAMPLE* input, int inputLength, WaveletCoeffs* coeffs, WaveletHistory* history) {
	coeffs->reset(inputLength, waveletFilterCoeffs[m_currWavelet].lowPassForwardLength);
	// check buffer
	assertWorkingBuffer(inputLength);
	if (m_algorithm==ALGORITHM_NUMERICAL_RECIPES) {
		numericalRecipesForwardTransform(input, inputLength, coeffs);
	} else {
		generalForwardTransform(input, inputLength, coeffs, history);
	}
}

void WaveletTransform::inverseTransform(WaveletCoeffs* input, SAMPLE* output) {
	// check buffer
	assertWorkingBuffer(input->getTotalSignalLength());
	if (m_algorithm==ALGORITHM_NUMERICAL_RECIPES) {
		numericalRecipesInverseTransform(input, output);
	} else {
		generalInverseTransform(input, output);
	}
}


int WaveletTransform::getWaveletCount() { 
	return getWaveletFilterCoeffsCount(); 
}

void WaveletTransform::setWavelet(int number) {
	if (number<0 || number>=getWaveletCount()) {
		AfxMessageBox("WaveletTransform::setWavelet(%d): out of range.");
//		EH_error(2, "WaveletTransform::setWavelet(%d): out of range.", number);
		return;
	}
	if (number==m_currWavelet) {
		return;
	}
	m_currWavelet=number;
	
	// set parameters for numerical recipes implementation
	m_wfilt.cc=waveletFilterCoeffs[m_currWavelet].lowPassFlipped-1;
	m_wfilt.cr=waveletFilterCoeffs[m_currWavelet].highPassFlipped-1;
	m_wfilt.ncof=waveletFilterCoeffs[m_currWavelet].highPassForwardLength;
	m_wfilt.ioff = m_wfilt.joff = 0;
	//wfilt.ioff = 2; wfilt.joff = -wfilt.ncof+2;

	// set parameters for my implementation
	setTransformFuncs();
}

char* WaveletTransform::getWaveletName() {
	return waveletFilterCoeffs[m_currWavelet].name;
}

int WaveletTransform::getAlgorithmCount() { 
	return ALGORITHM_COUNT;
}

void WaveletTransform::setTransformFuncs() {
	// set forward functions
	if (m_algorithm==ALGORITHM_ZERO_PAD_GENERIC) {
		m_forwardFunc=&forwardFilterZeroPadGeneric;
	} else
	if (m_algorithm==ALGORITHM_ZERO_PAD_OPTIMIZED) {
		m_forwardFunc=&forwardFilterZeroPadOptimized;
	} else
	if (m_algorithm==ALGORITHM_MIRROR_GENERIC) {
		m_forwardFunc=&forwardFilterMirrorGeneric;
	} else
	if (m_algorithm==ALGORITHM_ZERO_PAD_HISTORY_GENERIC) {
		m_forwardFunc=&forwardFilterHistoryGeneric;
	}
	// set inverse functions
	if (m_algorithm==ALGORITHM_ZERO_PAD_OPTIMIZED) {
		// optimized version
		m_inverseFunc=&inverseFilterOptimized;
	} else {
		// generic implementation
		m_inverseFunc=&inverseFilterGeneric;
	}
}

void WaveletTransform::setAlgorithm(int number) {
	if (number<0 || number>=getAlgorithmCount()) {
		AfxMessageBox("WaveletTransform::setAlgorithm(%d): out of range.");
//		EH_error(2, "WaveletTransform::setAlgorithm(%d): out of range.", number);
		return;
	}
	if (number==m_algorithm) {
		return;
	}
	m_algorithm=number;
	setTransformFuncs();
}

char* WaveletTransform::getAlgorithmName() {
	if (m_algorithm==ALGORITHM_ZERO_PAD_GENERIC) {
		return "Zero padding generic";
	} else
	if (m_algorithm==ALGORITHM_ZERO_PAD_OPTIMIZED) {
		return "Zero padding optimized";
	} else
	if (m_algorithm==ALGORITHM_MIRROR_GENERIC) {
		return "Mirrored padding generic";
	} else
	if (m_algorithm==ALGORITHM_ZERO_PAD_HISTORY_GENERIC) {
		return "Padding with history generic";
	} else
	if (m_algorithm=ALGORITHM_NUMERICAL_RECIPES) {
		return "Numerical Recipes (wrap)";
	}
	return "";
}

//////////////////// Numerical Recipes implementation /////////////////////////////////////

// NR routines

/* (C) Copr. 1986-92 Numerical Recipes Software W2,. */

void wt1(double a[], double worksp[], unsigned long n, int isign, wavefilt* wfilt,
		 void (*wtstep)(double [], double[], unsigned long, int, wavefilt*))
{
	unsigned long nn;

	if (n < 4) return;
	if (isign >= 0) {
		for (nn=n;nn>=4;nn>>=1) (*wtstep)(a, worksp, nn,isign, wfilt);
	} else {
		for (nn=4;nn<=n;nn<<=1) (*wtstep)(a, worksp, nn,isign, wfilt);
	}
}

void pwt(double a[], double worksp[], unsigned long n, int isign, wavefilt* wfilt)
{
	double ai,ai1;
	unsigned long i,ii,j,jf,jr,k,n1,ni,nj,nh,nmod;

	if (n < 4) return;
	nmod=wfilt->ncof*n;
	n1=n-1;
	nh=n >> 1;
	for (j=1;j<=n;j++) worksp[j]=0.0;
	if (isign >= 0) {
		for (ii=1,i=1;i<=n;i+=2,ii++) {
			ni=i+nmod+wfilt->ioff;
			nj=i+nmod+wfilt->joff;
			for (k=1;k<=(unsigned int) wfilt->ncof;k++) {
				jf=n1 & (ni+k);
				jr=n1 & (nj+k);
				worksp[ii] += wfilt->cc[k]*a[jf+1];		// low pass (approx)
				worksp[ii+nh] += wfilt->cr[k]*a[jr+1];	// high pass (details)
			}
		}
	} else {
		for (ii=1,i=1;i<=n;i+=2,ii++) {
			ai=a[ii];
			ai1=a[ii+nh];
			ni=i+nmod+wfilt->ioff;
			nj=i+nmod+wfilt->joff;
			for (k=1;k<=(unsigned int) wfilt->ncof;k++) {
				jf=(n1 & (ni+k))+1;
				jr=(n1 & (nj+k))+1;
				worksp[jf] += wfilt->cc[k]*ai;
				worksp[jr] += wfilt->cr[k]*ai1;
			}
		}
	}
	for (j=1;j<=n;j++) a[j]=worksp[j];
}

